import sklearn
import numpy as np

from . import EvaluationBase
from ..core.masks import BinaryMask


class PrecisionRecallFScore(EvaluationBase):
    """
    This class provides common statistical metrics for determining how well a source separation algorithm in nussl was
    able to create a binary mask compared to a known binary mask. The metrics used here are 
    `Precision, Recall <https://en.wikipedia.org/wiki/Precision_and_recall>`_,
    `F-Score <https://en.wikipedia.org/wiki/F1_score>`_ (sometimes called F-measure or F1-score), and Accuracy
    (though this is not reflected in the name of the class, it is simply   ``# correct / total``).

    Notes:
        * ``PrecisionRecallFScore`` can only be run using :ref:`binary_mask` objects. The constructor expects a list of 
        :ref:`binary_mask` objects for both the ground truth sources and the estimated sources.
        * ``PrecisionRecallFScore`` does not calculate the correct permutation of the estimated and ground truth sources;
        they are expected to be in the correct order when they are passed into ``PrecisionRecallFScore``.

    Args:
        true_sources_mask_list (list): List of :ref:`binary_mask` objects representing the ground truth sources.
        estimated_sources_mask_list (list): List of :ref:`binary_mask` objects representing the estimates from a source
         separation object
        source_labels (list) (Optional): List of ``str`` with labels for each source. If no labels are provided, sources
         will be labeled ``Source 0, Source 1, ...`` etc.        
    """

    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None,
                 compute_permutation=False, best_permutation_key="F1-Score", **kwargs):
        self.true_sources_list = self._verify_input_list(true_sources_list)
        self.estimated_sources_list = self._verify_input_list(estimated_sources_list)

        super().__init__(true_sources_list, estimated_sources_list,
                         source_labels=source_labels,
                         compute_permutation=compute_permutation,
                         best_permutation_key=best_permutation_key, **kwargs)
        self.keys = ['Precision', 'Recall', 'Accuracy', 'F1-Score']

    @staticmethod
    def _verify_input_list(mask_list):
        if isinstance(mask_list, BinaryMask):
            mask_list = [mask_list]
        elif isinstance(mask_list, list):
            if not all(isinstance(m, BinaryMask) for m in mask_list):
                raise ValueError('All items in mask_list must be of type nussl.BinaryMask!')

        if not all(mask_list[0].shape == m.shape for m in mask_list):
            raise ValueError('All masks must be the same shape!')

        return mask_list

    def preprocess(self):
        n_channels = self.true_sources_list[0].num_channels
        references = np.stack(
            [np.copy(x.mask.reshape(-1, n_channels))
             for x in self.true_sources_list],
            axis=-1
        )
        estimates = np.stack(
            [np.copy(x.mask.reshape(-1, n_channels))
             for x in self.estimated_sources_list],
            axis=-1
        )
        return references, estimates

    def evaluate_helper(self, references, estimates, **kwargs):
        """
        Determines the precision, recall, f-score, and accuracy of each :ref:`binary_mask` object in 
        ``true_sources_mask_list`` and ``estimated_sources_mask_list``. Returns a list of results that is
        formatted like so:
        
        .. code-block:: python

            [              
                {'Accuracy': 0.83,
                'Precision': 0.78,
                'Recall': 0.81,
                'F1-Score': 0.77 },

                {'Accuracy': 0.22,
                'Precision': 0.12,
                'Recall': 0.15,
                'F1-Score': 0.19 }
            ]
                
        Returns:
            self.scores (dict): A list of scores that contains accuracy, precision, recall, and F1-score
            of between the list of :ref:`binary_mask` objects in both ``true_sources_mask_list`` 
            and ``estimated_sources_mask_list``.

        """
        scores = []
        for j in range(references.shape[-1]):
            ch_metrics = {
                'Accuracy': [], 'Precision': [], 'Recall': [], 'F1-Score': []
            }
            for ch in range(references.shape[-2]):
                ch_metrics['Accuracy'].append(sklearn.metrics.accuracy_score(
                    references[..., ch, j], estimates[..., ch, j]
                ))
                ch_metrics['Precision'].append(sklearn.metrics.precision_score(
                    references[..., ch, j], estimates[..., ch, j]
                ))
                ch_metrics['Recall'].append(sklearn.metrics.recall_score(
                    references[..., ch, j], estimates[..., ch, j]
                ))
                ch_metrics['F1-Score'].append(sklearn.metrics.f1_score(
                    references[..., ch, j], estimates[..., ch, j]
                ))
            scores.append(ch_metrics)

        return scores



"""
While *nussl* does not come with any data sets, it does have the capability to interface with
many common source separation data sets used within the MIR and speech separation communities.
These data set "hooks" subclass BaseDataset and by default return AudioSignal objects in
labeled dictionaries for ease of use. Transforms can be applied to these datasets for use
in machine learning pipelines.
"""
import os

from .. import musdb
import jams

from ..core import constants, utils
from .base_dataset import BaseDataset, DataSetException


class MUSDB18(BaseDataset):
    """
    Hook for MUSDB18. Uses the musdb.DB object to access the
    dataset. If ``download=True``, then the 7s snippets of each track
    are downloaded to ``self.folder``. If no folder is given, then
    the tracks are downloaded to ~/.nussl/musdb18. 
    
    Getting an item from this dataset with no transforms returns the 
    following dictionary:

    .. code-block:: none

        {
            'mix': [AudioSignal object containing mix audio],
            'source': {
                'bass': [AudioSignal object containing vocals],
                'drums': [AudioSignal object containing drums],
                'other': [AudioSignal object containing other],
                'vocals': [AudioSignal object containing vocals],
            }
            'metadata': {
                'labels': ['bass', 'drums', 'other', 'vocals']
            }
        }
    
    Args:
        folder (str, optional): Location that should be processed to produce the 
            list of files. Defaults to None.
        is_wav (bool, optional):  Expect subfolder with wav files for each source 
            instead of stems, defaults to False.
        download (bool, optional): Download sample version of MUSDB18 which 
            includes 7s excerpts. Defaults to False.
        subsets (list, optional): Select a musdb subset train or test. 
            Defaults to ['train', 'test'] (all tracks).
        split (str, optional): When subset train is loaded, split selects the 
            train/validation split. split=’train’ loads the training split, 
            `split=’valid’ loads the validation split. split=None applies no 
            splitting. Defaults to None.
        **kwargs: Any additional arguments that are passed up to BaseDataset 
            (see ``nussl.datasets.BaseDataset``).
    """
    DATASET_HASHES = {
        "musdb": "56777516ad56fe6a8590badf877e6be013ff932c010e0fbdb0aba03ef878d4cd",
    }
    
    def __init__(self, folder=None, is_wav=False, download=False,
                 subsets=None, split=None, **kwargs):
        subsets = ['train', 'test'] if subsets is None else subsets
        if folder is None:
            folder = os.path.join(
                constants.DEFAULT_DOWNLOAD_DIRECTORY, 'musdb18'
            )
        self.musdb = musdb.DB(root=folder, is_wav=is_wav, download=download, 
                              subsets=subsets, split=split)
        super().__init__(folder, **kwargs)
        self.metadata['subsets'] = subsets
        self.metadata['split'] = split

    def get_items(self, folder):
        items = range(len(self.musdb))
        return list(items)

    def process_item(self, item):
        track = self.musdb[item]
        mix, sources = utils.musdb_track_to_audio_signals(track)
        self._setup_audio_signal(mix)
        for source in list(sources.values()):
            self._setup_audio_signal(source)
        
        output = {
            'mix': mix,
            'sources': sources,
            'metadata': {
                'labels': ['bass', 'drums', 'other', 'vocals']
            }
        }
        return output


class MixSourceFolder(BaseDataset):
    """
    This dataset expects your data to be formatted in the following way:

    .. code-block:: none

        data/
            mix/
                [file0].wav
                [file1].wav
                [file2].wav
                ...
            [label0]/
                [file0].wav
                [file1].wav
                [file2].wav
                ...
            [label1]/
                [file0].wav
                [file1].wav
                [file2].wav
                ...
            [label2]/
                [file0].wav
                [file1].wav
                [file2].wav
                ...
            ...

    Note that the the filenames match between the mix folder and each source folder.
    The source folder names can be whatever you want. Given a file in the 
    ``self.mix_folder`` folder, this dataset will look up the corresponding files 
    with the same name in the source folders. These are the source audio files. 
    The sum of the sources should equal the mixture. Each source will be labeled 
    according to the folder name it comes from.

    Getting an item from this dataset with no transforms returns the 
    following dictionary:

    .. code-block:: none

        {
            'mix': [AudioSignal object containing mix audio],
            'source': {
                '[label0]': [AudioSignal object containing label0 audio],
                '[label1]': [AudioSignal object containing label1 audio],
                '[label2]': [AudioSignal object containing label2 audio],
                '[label3]': [AudioSignal object containing label3 audio],
                ...
            }
            'metadata': {
                'labels': ['label0', 'label1', 'label2', 'label3']
            }
        }


    Args:
        folder (str, optional): Location that should be processed to produce the 
            list of files. Defaults to None.
        mix_folder (str, optional): Folder to look in for mixtures. Defaults to 'mix'.
        source_folders (list, optional): List of folders to look in for sources. 
            Path is defined relative to folder. If None, all folders other than 
            mix_folder are treated as the source folders. Defaults to None.
        ext (list, optional): Audio extensions to look for in mix_folder. 
            Defaults to ['.wav', '.flac', '.mp3'].
        **kwargs: Any additional arguments that are passed up to BaseDataset 
            (see ``nussl.datasets.BaseDataset``).
    """
    def __init__(self, folder, mix_folder='mix', source_folders=None,
                 ext=None, make_mix=False, **kwargs):
        self.mix_folder = mix_folder
        self.source_folders = source_folders
        self.ext = ['.wav', '.flac', '.mp3'] if ext is None else ext
        self.make_mix = make_mix
        super().__init__(folder, **kwargs)

    def get_items(self, folder):
        if self.source_folders is None:
            self.source_folders = sorted([
                f for f in os.listdir(folder)
                if os.path.isdir(os.path.join(folder, f))
                and f != self.mix_folder
            ])

        if self.make_mix:
            mix_folder = os.path.join(folder, self.source_folders[0])
        else:
            mix_folder = os.path.join(folder, self.mix_folder)
        items = sorted([
            x for x in os.listdir(mix_folder)
            if os.path.splitext(x)[1] in self.ext
        ])
        return items

    def get_mix_and_sources(self, item):
        sources = {}
        for k in self.source_folders:
            source_path = os.path.join(self.folder, k, item)
            if os.path.exists(source_path):
                sources[k] = self._load_audio_file(source_path)
        
        if self.make_mix:
            mix = sum(list(sources.values()))
        else:
            mix_path = os.path.join(self.folder, self.mix_folder, item)
            mix = self._load_audio_file(mix_path)
        return mix, sources

    def process_item(self, item):
        mix, sources = self.get_mix_and_sources(item)
        output = {
            'mix': mix,
            'sources': sources,
            'metadata': {
                'labels': self.source_folders
            }
        }
        return output

class Scaper(BaseDataset):
    """
    Source separation datasets can be generated using Scaper, a library for
    automatic soundscape generation. Datasets that are generated with Scaper
    can be fed into this class easily. Scaper generates a large list of JAMS
    files which specify the parameters of the soundscape. If the soundscape is
    generated with `save_isolated_events=True`, then the audio corresponding
    to each event in the soundscape will be saved as well.

    Below is an example of using Scaper to generate a small dataset of 10 
    mixtures with 2 sources each. The generated dataset can then be immediately
    loaded into an instance of ``nussl.datasets.Scaper`` for integration into
    a training or evaluation pipeline.

    The sources are output in a dictionary that looks like this:

    .. code-block:: none

        data['sources] = {
            '{label}::{count}': AudioSignal,
            '{label}::{count}': AudioSignal,
            ...
        }

    For example:

    .. code-block:: none

        data['sources] = {
            'siren::0': AudioSignal,
            'siren::1': AudioSignal,
            'car_horn::0': AudioSignal,
            ...
        }

    Getting an item from this dataset with no transforms returns the 
    following dictionary:

    .. code-block:: none

        {
            'mix': [AudioSignal object containing mix audio],
            'source': {
                '[label0::count]': [AudioSignal object containing label0 audio],
                '[label1::count]': [AudioSignal object containing label1 audio],
                '[label2::count]': [AudioSignal object containing label2 audio],
                '[label3::count]': [AudioSignal object containing label3 audio],
                ...
            }
            'metadata': {
                'jams': [the content of the jams file used to generate the soundscape]
                'labels': ['label0', 'label1', 'label2', 'label3']
            }
        }


    Example of generating a Scaper dataset and then loading it with nussl:

    >>> n_sources = 2
    >>> n_mixtures = 10
    >>> duration = 3
    >>> ref_db = -40
    >>> fg_path = '/path/to/foreground/'
    >>> output_dir = '/output/path'
    >>> for i in range(n_mixtures):
    >>>     sc = scaper.Scaper(
    >>>         duration, fg_path, fg_path, random_state=i)
    >>>     sc.ref_db = ref_db
    >>>     sc.sr = 16000
    >>>     for j in range(n_sources):
    >>>         sc.add_event(
    >>>             label=('choose', []),
    >>>             source_file=('choose', []),
    >>>             source_time=('const', 0),
    >>>             event_time=('const', 0),
    >>>             event_duration=('const', duration),
    >>>             snr=('const', 0),
    >>>             pitch_shift=None,
    >>>             time_stretch=None
    >>>         )
    >>>     audio_path = os.path.join(output_dir, f'{i}.wav')
    >>>     jams_path = os.path.join(output_dir, f'{i}.jams')
    >>>     sc.generate(audio_path, jams_path, save_isolated_events=True)
    >>> dataset = nussl.datasets.Scaper(output_dir)
    >>> dataset[0] # contains mix, sources, and metadata corresponding to 0.jams.
        
    Raises:
        DataSetException: if Scaper dataset wasn't saved with isolated event audio.
    """
    def get_items(self, folder):
        items = sorted([
            x for x in os.listdir(folder)
            if os.path.splitext(x)[1] in ['.jams']
        ])
        return items

    def _get_info_from_item(self, item):
        jam = jams.load(os.path.join(self.folder, item))
        ann = jam.annotations.search(namespace='scaper')[0]
        mix_path = ann.sandbox.scaper['soundscape_audio_path']
        source_paths = ann.sandbox.scaper['isolated_events_audio_path']
        return jam, ann, mix_path, source_paths

    def process_item(self, item):
        jam, ann, mix_path, source_paths = self._get_info_from_item(item)
        if not source_paths:
            raise DataSetException(
                "No paths to isolated events found! Did you generate "
                "the soundscape with save_isolated_events=True?")

        mix = self._load_audio_file(mix_path)
        sources = {}

        for event_spec, event_audio_path in zip(ann, source_paths):
            label = event_spec.value['label']
            label_count = 0
            for k in sources:
                if label in k:
                    label_count += 1
            label = f"{label}::{label_count}"
            sources[label] = self._load_audio_file(event_audio_path)

        output = {
            'mix': mix,
            'sources': sources,
            'metadata': {
                'scaper': jam,
                'labels': ann.sandbox.scaper['fg_labels'],
            }
        }
        return output

class OnTheFly(BaseDataset):
    """
    Hook for a dataset that creates mixtures on the fly from source
    data. The function that creates the mixture is a closure which
    is defined by the end-user. The number of mixtures in the 
    dataset is also defined by the end-user. The mix closure function
    should take two arguments - the dataset object and the index of the 
    item being processed - and the output of the mix closure should be a 
    dictionary containing at least a 'mix', 'sources' and (optionally) 
    a 'metadata' key, or other keys that can be defined up to you. 
    Here's an example of a closure, which can be configured via 
    variable scoping:
    
    >>>  def make_sine_wave(freq, sample_rate, duration):
    >>>      dt = 1 / sample_rate
    >>>      x = np.arange(0.0, duration, dt)
    >>>      x = np.sin(2 * np.pi * freq * x)
    >>>      return x
    >>>  n_sources = 2
    >>>  duration = 3
    >>>  sample_rate = 44100
    >>>  min_freq, max_freq = 110, 1000
    >>>  def make_mix(dataset, i):
    >>>      sources = {}
    >>>      freqs = []
    >>>      for i in range(n_sources):
    >>>          freq = np.random.randint(min_freq, max_freq)
    >>>          freqs.append(freq)
    >>>          source_data = make_sine_wave(freq, sample_rate, duration)
    >>>          source_signal = dataset._load_audio_from_array(
    >>>              audio_data=source_data, sample_rate=sample_rate)
    >>>          sources[f'sine{i}'] = source_signal * 1 / n_sources
    >>>      mix = sum(sources.values())
    >>>      output = {
    >>>          'mix': mix,
    >>>          'sources': sources,
    >>>          'metadata': {
    >>>              'frequencies': freqs    
    >>>          }    
    >>>      }
    >>>      return output
    >>>  dataset = nussl.datasets.OnTheFly(make_mix, 10)

    Args:
        mix_closure (function): A closure that determines how to create
          a single mixture, given the index. It has a strict input 
          signature (the index is given as an int) and a strict output
          signature (a dictionary containing a 'mix' and 'sources') key.
        num_mixtures (int): Number of mixtures that will be created on
          the fly. This determines one 'run' thrugh the dataset, or an 
          epoch.
        kwargs: Keyword arguments to BaseDataset.
    """
    def __init__(self, mix_closure, num_mixtures, **kwargs):
        self.num_mixtures = num_mixtures
        self.mix_closure = mix_closure

        super().__init__('none', **kwargs)
        self.metadata['num_mixtures'] = num_mixtures

    def get_items(self, folder):
        return list(range(self.num_mixtures))
    
    def process_item(self, item):
        output = self.mix_closure(self, item)
        if not isinstance(output, dict):
            raise DataSetException("output of mix_closure must be a dict!")
        if 'mix' not in output or 'sources' not in output:
            raise DataSetException(
                "output of mix_closure must be a dict containing "
                "'mix', 'sources' as keys!")
        return output

class FUSS(Scaper):
    """
    The Free Universal Sound Separation (FUSS) Dataset is a database of arbitrary 
    sound mixtures and source-level references, for use in experiments on 
    arbitrary sound separation. 

    This is the official sound separation data for the DCASE2020 Challenge Task 4: 
    Sound Event Detection and Separation in Domestic Environments.

    This is a hook for reading in this dataset, and making sure that the mix and 
    source paths are massaged to be relative paths.

    References:

    [1]  Scott Wisdom, Hakan Erdogan, Daniel P. W. Ellis, Romain Serizel, 
    Nicolas Turpault, Eduardo Fonseca, Justin Salamon, Prem Seetharaman, 
    John R. Hershey, "What's All the FUSS About Free Universal Sound Separation 
    Data?", 2020, in preparation.

    [2] Eduardo Fonseca, Jordi Pons, Xavier Favory, Frederic Font Corbera, 
    Dmitry Bogdanov, Andrés Ferraro, Sergio Oramas, Alastair Porter, and 
    Xavier Serra. "Freesound Datasets: A Platform for the Creation of Open Audio 
    Datasets." International Society for Music Information Retrieval Conference 
    (ISMIR), pp. 486–493. Suzhou, China, 2017.
    
    Args:
        root (str): Folder where the FUSS data is. Either points to ssdata or 
          ssdata_reverb.
        split (str): Either the ``train``, ``validation``, or ``eval`` split. 
        kwargs: Additional keyword arguments to BaseDataset.
    """
    def __init__(self, root, split='train', **kwargs):
        if split not in ['train', 'validation', 'eval']:
            raise DataSetException(
                f"split '{split}' not one of the accepted splits: "
                f"'train', 'validation', 'eval'.")
        
        folder = os.path.join(root, split)
        super().__init__(folder, sample_rate=16000, strict_sample_rate=True, 
                         **kwargs)
        self.metadata['split'] = split

    def _get_info_from_item(self, item):
        path_to_item = os.path.join(self.folder, item)
        item_base_name = os.path.splitext(item)[0]

        jam = jams.load(path_to_item)
        ann = jam.annotations.search(namespace='scaper')[0]
        mix_path = ann.sandbox.scaper['soundscape_audio_path']
        source_paths = ann.sandbox.scaper['isolated_events_audio_path']

        mix_path = os.path.join(
            self.folder, item_base_name + mix_path.split(item_base_name)[-1])
        for i, source_path in enumerate(source_paths):
            source_paths[i] = os.path.join(
                self.folder, item_base_name + source_path.split(item_base_name)[-1])

        return jam, ann, mix_path, source_paths


class WHAM(MixSourceFolder):
    """
    Hook for the WHAM dataset. Essentially subclasses MixSourceFolder but with presets
    that are helpful for WHAM, which as the following directory structure:

    .. code-block:: none

        [wav8k, wav16k]/
          [min, max]/
            [tr, cv, tt]/
                mix_both/
                mix_clean/
                mix_single/
                noise/
                s1/
                s2/
        wham_noise/
          tr/
          cv/
          tt/
          metadata/

    Args:
        root (str): Root of WHAM directory.
        mix_folder (str): Which folder is the mix? Either 'mix_clean', 'mix_both', or
          'mix_single'.
        mode (str): Either 'min' or 'max' mode.
        split (str): Split to use (tr, cv, or tt).
        sample_rate (int): Sample rate of audio, either 8000 or 16000.
    """
    MIX_TO_SOURCE_MAP = {
        'mix_clean': ['s1', 's2'],
        'mix_both': ['s1', 's2', 'noise'],
        'mix_single': ['s1'],
    }

    DATASET_HASHES = {
        "wav8k": "acd49e0dae066e16040c983d71cc5a8adb903abff6e5cbb92b3785a1997b7547", 
        "wav16k": "5691d6a35382f2408a99594f21d820b58371b5ea061841db37d548c0b8d6ec7f"
    }

    def __init__(self, root, mix_folder='mix_clean', mode='min', split='tr', 
                 sample_rate=8000, **kwargs):
        if mix_folder not in self.MIX_TO_SOURCE_MAP.keys():
            raise DataSetException(
                f"{mix_folder} must be in {list(self.MIX_TO_SOURCE_MAP.keys())}")
        if sample_rate not in [8000, 16000]:
            raise DataSetException(
                f"{sample_rate} not available for WHAM (only 8000 and 16000 Hz allowed)")
        if mode not in ['min', 'max']:
            raise DataSetException(
                f"{mode} not available, only 'min' or 'max' allowed.")
        if split not in ['tr', 'cv', 'tt']:
            raise DataSetException(
                f"{split} not available, must be one of 'tr' (train), "
                f"'cv' (validation), and 'tt' (test)")

        wav_folder = 'wav8k' if sample_rate == 8000 else 'wav16k'
        folder = os.path.join(root, wav_folder, mode, split)
        source_folders = self.MIX_TO_SOURCE_MAP[mix_folder]

        super().__init__(folder, mix_folder=mix_folder, source_folders=source_folders,
                         sample_rate=sample_rate, strict_sample_rate=True, **kwargs)
        self.metadata.update({
            'mix_folder': mix_folder,
            'mode': mode,
            'split': split,
            'wav_folder': wav_folder
        })



import copy

import torch

from . import loss
from .trainer import BackwardsEvents


class Closure(object):
    """
    Closures are used with ignite Engines to train a model given an optimizer 
    and a set of loss functions. Closures perform forward passes of models given 
    the input data. The loss is computed via ``self.compute_loss``. 
    The forward pass is implemented via the objects ``__call__`` function. 

    This closure object provides a way to define the loss functions you want to
    use to train your model as a loss dictionary that is structured as follows:

    .. code-block:: python

        loss_dictionary = {
            'LossClassName': {
                'weight': [how much to weight the loss in the sum, defaults to 1],
                'keys': [key mapping items in dictionary to arguments to loss],
                'args': [any positional arguments to the loss class],
                'kwargs': [keyword arguments to the loss class],
            }
        }

    The keys value will default to ``LossClassName.DEFAULT_KEYS``, which can be
    found in ``nussl.ml.train.loss`` within each available class. Here's an example
    of a Chimera loss combining deep clustering with permutation invariant L1 loss:

    .. code-block:: python 

        loss_dictionary = {
            'DeepClusteringLoss': {
                'weight': .2,
            },
            'PermutationInvariantLoss': {
                'weight': .8,
                'args': ['L1Loss']
            }
        }

    Or if you're using permutation invariant loss but need to specify arguments to the
    loss function being wrapped by PIT, you can do this:

    .. code-block:: python

        loss_dictionary = {
            'PITLoss': {
                'class': 'PermutationInvariantLoss',
                'keys': {'audio': 'estimates', 'source_audio': 'targets'},
                'args': [{
                    'class': 'SISDRLoss',
                    'kwargs': {'scaling': False}
                }]
            }
        }


    If you have your own loss function classes you wish to use, you can pass those
    into the loss dictionary and make them discoverable by the closure by using
    `ml.register_loss.`

    Args:
        loss_dictionary (dict): Dictionary of losses described above.
        combination_approach (str): How to combine losses, if there are multiple
         losses. The default is that the losses will be combined via a weighted
         sum ('combine_by_sum'). Can also do 'combine_by_multiply'. Defaults to
         'combine_by_sum'.
        args: Positional arguments to ``combination_approach``.
        kwargs: Keyword arguments to ``combination_approach``.

    See also:
        ml.register_loss to register your loss functions with this closure.
    """

    def __init__(self, loss_dictionary, combination_approach='combine_by_sum', 
                 *args, **kwargs):
        loss_dictionary = self._validate_loss_dictionary(loss_dictionary)

        self.combination_func = getattr(self, combination_approach)
        self.args = args
        self.kwargs = kwargs

        self.losses = []
        for key, val in loss_dictionary.items():
            _loss_name = val['class'] if 'class' in val else key
            loss_class = getattr(loss, _loss_name)
            weight = 1 if 'weight' not in val else val['weight']
            keys = loss_class.DEFAULT_KEYS if 'keys' not in val else val['keys']
            args = [] if 'args' not in val else copy.deepcopy(val['args'])
            kwargs = {} if 'kwargs' not in val else copy.deepcopy(val['kwargs'])
            if _loss_name in ['CombinationInvariantLoss', 'PermutationInvariantLoss']:
                if isinstance(args[0], str):
                    args[0] = getattr(loss, args[0])()
                elif isinstance(args[0], dict):
                    arg_class = getattr(loss, args[0]['class'])
                    args_to_loss = [] if 'args' not in args[0] else args[0]['args']
                    kwargs_to_loss = {} if 'kwargs' not in args[0] else args[0]['kwargs']
                    args[0] = arg_class(*args_to_loss, **kwargs_to_loss)

            _loss = (loss_class(*args, **kwargs), weight, keys, key)
            self.losses.append(_loss)

    @staticmethod
    def _validate_loss_dictionary(loss_dictionary):
        if not isinstance(loss_dictionary, dict):
            raise ClosureException(
                "loss_dictionary must be a dictionary specifying the "
                "class and arguments for each loss function! ")

        for key, val in loss_dictionary.items():
            _loss = val['class'] if 'class' in val else key
            if _loss not in dir(loss):
                raise ClosureException(
                    f"Loss function {_loss} not found in loss which has {dir(loss)}")

            if not isinstance(val, dict):
                raise ClosureException(
                    "Each key in loss dictionary must point to a dict!")

            for val_key in val:
                if val_key not in ['weight', 'keys', 'args', 'kwargs', 'class']:
                    raise ClosureException(
                        f"{key} in loss_dictionary not in ['weight', 'args', 'kwargs'")

                elif val_key == 'weight':
                    if not isinstance(val[val_key], (float, int)) and not torch.is_tensor(val[val_key]):
                        raise ClosureException(f"weight can only be an int or a float")

                elif val_key == 'args':
                    if not isinstance(val[val_key], list):
                        raise ClosureException(f"args must be a list")

                elif val_key == 'kwargs':
                    if not isinstance(val[val_key], dict):
                        raise ClosureException("kwargs must be a dict")
        
        return loss_dictionary

    def __call__(self, engine, data):
        raise NotImplementedError()

    def combine_by_multitask(self, loss_output):
        """
        Implements a multitask learning objective [1] where each loss
        is weighted by a learned parameter with the following
        function:

        combined_loss = \sum_i exp(-weight_i) * loss_i + weight_i

        where i indexes each loss. The weights come from the loss 
        dictionary and can point to nn.Parameter teensors that get 
        learned jointly with the model.

        References:

        [1] Kendall, Alex, Yarin Gal, and Roberto Cipolla. 
            "Multi-task learning using uncertainty to weigh losses 
            for scene geometry and semantics." Proceedings of the 
            IEEE conference on computer vision and pattern recognition. 2018.
        """
        combined_loss = 0
        for _, weight, _, name in self.losses:
            sigma = torch.exp(-weight)
            combined_loss += sigma * loss_output[name] + weight
        return combined_loss

    def combine_by_multiply(self, loss_output):
        combined_loss = 1
        for _, weight, _, name in self.losses:
            combined_loss *= weight * loss_output[name]
        return combined_loss

    def combine_by_sum(self, loss_output):
        combined_loss = 0
        for _, weight, _, name in self.losses:
            # if the weight is 0, then the loss is just for
            # monitoring and we won't bother summing with it, 
            # in case its shape doesnt match.
            if weight != 0:
                combined_loss += weight * loss_output[name]
        return combined_loss

    def compute_loss(self, output, target):
        loss_output = {}
        output.update(target)

        for loss_obj, weight, keys, name in self.losses:
            kwargs = {keys[k]: output[k] for k in keys}
            loss_output[name] = loss_obj(**kwargs)
        
        loss_output['loss'] = self.combination_func(
            loss_output, *self.args, **self.kwargs)
        return loss_output


class TrainClosure(Closure):
    """
    This closure takes an optimization step on a SeparationModel object given a
    loss.
    
    Args:
        loss_dictionary (dict): Dictionary containing loss functions and specification.
        optimizer (torch Optimizer): Optimizer to use to train the model.
        model (SeparationModel): The model to be trained.
    """

    def __init__(self, loss_dictionary, optimizer, model, *args, **kwargs):
        super().__init__(loss_dictionary, *args, **kwargs)
        self.optimizer = optimizer
        self.model = model

        # Save about training metadata to model.info
        self.model.metadata['optimizer'] = {
            'name': type(optimizer).__name__,
            'params': optimizer.defaults  # All of the settings are stored here.
        }
        self.model.metadata['loss_dictionary'] = loss_dictionary

    def _fire_event(self, engine, output, event):
        if engine is not None:
            if engine.state is not None:
                engine.state.model_output = output
            engine.fire_event(event)

    def __call__(self, engine, data):
        self.model.train()
        self.optimizer.zero_grad()

        output = self.model(data)

        loss_ = self.compute_loss(output, data)
        loss_['loss'].backward()
        self._fire_event(engine, output, BackwardsEvents.BACKWARDS_COMPLETED)
        self.optimizer.step()
        loss_ = {key: loss_[key].item() for key in loss_}

        return loss_


class ValidationClosure(Closure):
    """
    This closure validates the model on some data dictionary.
    
    Args:
        loss_dictionary (dict): Dictionary containing loss functions and specification.
        model (SeparationModel): The model to be validated.
    """

    def __init__(self, loss_dictionary, model, *args, **kwargs):
        super().__init__(loss_dictionary, *args, **kwargs)
        self.model = model

    def __call__(self, engine, data):
        with torch.no_grad():
            self.model.eval()
            output = self.model(data)
            loss_ = self.compute_loss(output, data)
            loss_ = {key: loss_[key].item() for key in loss_}
        return loss_


class ClosureException(Exception):
    """ 
    Exception class for errors when working with closures in nussl.
    """
    pass



import numpy as np

from ..base import ClusteringSeparationBase


class SpatialClustering(ClusteringSeparationBase):
    """
    Implements clustering on IPD/ILD features between the first two channels.

    IPD/ILD features are inter-phase difference and inter-level difference
    features. Sounds coming from different directions will naturally cluster
    in IPD/ILD space. 

    Subclasses ClusteringSeparationBase which actually handles all of the
    clustering functionality behind this function.
    """
    def extract_features(self):
        ipd, ild = self.audio_signal.ipd_ild_features()
        num_channels = self.audio_signal.num_channels
        
        features = [
            [ipd for _ in range(num_channels)],
            [ild for _ in range(num_channels)]
        ]
        features = np.array(features).transpose(2, 3, 1, 0)   

        return features



"""
Datasets
========

Base class
----------
.. autoclass:: nussl.datasets.BaseDataset
    :members:
    :autosummary:

MUSDB18
-------
.. autoclass:: nussl.datasets.MUSDB18
    :members:
    :autosummary:

WHAM
----
.. autoclass:: nussl.datasets.WHAM
    :members:
    :autosummary:

FUSS
----
.. autoclass:: nussl.datasets.FUSS
    :members:
    :autosummary:

Scaper
------
.. autoclass:: nussl.datasets.Scaper
    :members:
    :autosummary:

MixSourceFolder
---------------
.. autoclass:: nussl.datasets.MixSourceFolder
    :members:
    :autosummary:

OnTheFly
--------
.. autoclass:: nussl.datasets.OnTheFly
    :members:
    :autosummary:

Data transforms
---------------
.. automodule:: nussl.datasets.transforms
    :members:
    :autosummary:

"""

from .base_dataset import BaseDataset
from .hooks import (
    MUSDB18, 
    MixSourceFolder, 
    Scaper, 
    WHAM, 
    FUSS, 
    OnTheFly
)
from . import transforms



try:
    import vamp
    vamp_imported = True
except Exception:
    vamp_imported = False

# Current nussl version
__version__ = '1.1.3rc4'

class ImportErrorClass(object):
    def __init__(self, lib, **kwargs):
        raise ImportError(
            f'Cannot import {type(self).__name__} because {lib} is not installed')

import warnings

try:
    import musdb
except RuntimeError:
    warnings.warn(
        "Importing musdb failed, likely because ffmpeg or ffprobe are not installed!.")
    musdb = None

from .core import AudioSignal, STFTParams
from .core import utils, efz_utils, play_utils, constants, mixing


from . import core
from . import evaluation
from . import datasets
from . import ml
from . import separation




"""
Cluster sources by timbre
-------------------------

.. autoclass:: nussl.separation.primitive.TimbreClustering
    :autosummary:

Foreground/background via 2DFT
------------------------------

.. autoclass:: nussl.separation.primitive.FT2D
    :autosummary:

Harmonic/percussive separation
------------------------------

.. autoclass:: nussl.separation.primitive.HPSS
    :autosummary:

Foreground/background via REPET
-------------------------------

.. autoclass:: nussl.separation.primitive.Repet
    :autosummary:

Foreground/background via REPET-SIM
-----------------------------------

.. autoclass:: nussl.separation.primitive.RepetSim
    :autosummary:

Vocal melody extraction via Melodia
-----------------------------------

.. autoclass:: nussl.separation.primitive.Melodia
    :autosummary:

"""

from .timbre import TimbreClustering
from .ft2d import FT2D
from .hpss import HPSS
from .repet import Repet
from .repet_sim import RepetSim
from .melodia import Melodia



"""
These are optional utilities included in nussl that allow one to embed an AudioSignal
as a playable object in a Jupyter notebook, or to play audio from
the terminal.
"""
from copy import deepcopy
import subprocess
from tempfile import NamedTemporaryFile

import random, string
import importlib_resources as pkg_resources

from . import templates
from .utils import _close_temp_files

multitrack_template = pkg_resources.read_text(templates, 'multitrack.html')


def _check_imports():
    try:
        import ffmpy
    except:
        ffmpy = False

    try:
        import IPython
    except:
        raise ImportError('IPython must be installed in order to use this function!')
    return ffmpy, IPython


def embed_audio(audio_signal, ext='.mp3', display=True):
    """
    Write a numpy array to a temporary mp3 file using ffmpy, then embeds the mp3
    into the notebook.

    Args:
        audio_signal (AudioSignal): AudioSignal object containing the data.
        ext (str): What extension to use when embedding. '.mp3' is more lightweight 
          leading to smaller notebook sizes. Defaults to '.mp3'.
        display (bool): Whether or not to display the object immediately, or to return
          the html object for display later by the end user. Defaults to True.

    Example:
        >>> import nussl
        >>> audio_file = nussl.efz_utils.download_audio_file('schoolboy_fascination_excerpt.wav')
        >>> audio_signal = nussl.AudioSignal(audio_file)
        >>> audio_signal.embed_audio()

    This will show a little audio player where you can play the audio inline in 
    the notebook.      
    """
    audio_signal = deepcopy(audio_signal)
    ffmpy, IPython = _check_imports()
    sr = audio_signal.sample_rate
    tmpfiles = []

    with _close_temp_files(tmpfiles):
        tmp_wav = NamedTemporaryFile(
            mode='w+', suffix='.wav', delete=False)
        tmpfiles.append(tmp_wav)
        audio_signal.write_audio_to_file(tmp_wav.name)
        if ext != '.wav' and ffmpy:
            tmp_converted = NamedTemporaryFile(
                mode='w+', suffix=ext, delete=False)
            tmpfiles.append(tmp_wav)
            ff = ffmpy.FFmpeg(
                inputs={tmp_wav.name: None},
                outputs={tmp_converted.name: '-write_xing 0 -codec:a libmp3lame -b:a 128k -y'})
            ff.run()
        else:
            tmp_converted = tmp_wav

        audio_element = IPython.display.Audio(data=tmp_converted.name, rate=sr)
        if display:
            IPython.display.display(audio_element)
    return audio_element


def multitrack(audio_signals, names=None, ext='.mp3', display=True):
    """
    Takes a bunch of audio sources, converts them to mp3 to make them smaller, and
    creates a multitrack audio player in the notebook that lets you
    toggle between the sources and the mixture. Heavily adapted
    from https://github.com/binarymind/multitrackHTMLPlayer,
    designed by Bastien Liutkus.

    Args:
        audio_signals (list): List of AudioSignal objects that add up to the mixture.
        names (list): List of names to give to each object (e.g. foreground, background).
        ext (str): What extension to use when embedding. '.mp3' is more lightweight
          leading to smaller notebook sizes.
        display (bool): Whether or not to display the object immediately, or to return
          the html object for display later by the end user.
    """
    ffmpy, IPython = _check_imports()
    div_id = ''.join(random.choice(string.ascii_uppercase) for _ in range(20))
    _names = None

    if isinstance(audio_signals, dict):
        _names = list(audio_signals.keys())
        audio_signals = [audio_signals[k] for k in _names]

    if names is not None:
        if len(names) != len(audio_signals):
            raise ValueError("len(names) must be equal to len(audio_signals)!")
    else:
        if _names is not None:
            names = _names
        else:
            names = [
                f"{i}:{s.path_to_input_file}"
                for i, s in enumerate(audio_signals)
            ]

    template = (
        f"<div id={div_id} class=audio-container "
        f"preload=auto name={div_id}>")

    for name, signal in zip(names, audio_signals):
        encoded_audio = embed_audio(signal, ext=ext, display=False).src_attr()
        audio_element = (
            f"<audio name='{name}' url={encoded_audio}></audio>")
        template += audio_element

    template += "</div>"
    template += multitrack_template
    template = template.replace('NAME', div_id)

    html = IPython.display.HTML(template)
    if display:
        IPython.display.display(html)
    else:
        return html


def play(audio_signal):
    """
    Plays an audio signal if ffplay from the ffmpeg suite of tools is installed.
    Otherwise, will fail. The audio signal is written to a temporary file
    and then played with ffplay.
    
    Args:
        audio_signal (AudioSignal): AudioSignal object to be played.
    """
    tmpfiles = []
    with _close_temp_files(tmpfiles):
        tmp_wav = NamedTemporaryFile(suffix='.wav', delete=False)
        tmpfiles.append(tmp_wav)
        audio_signal.write_audio_to_file(tmp_wav.name)
        subprocess.call(["ffplay", "-nodisp", "-autoexit", tmp_wav.name])



import numpy as np

from .. import ClusteringSeparationBase, SeparationException


class EnsembleClustering(ClusteringSeparationBase):
    """
    Run multiple separation algorithms on a single mixture and concatenate their
    masks to input into a clustering algorithm.
    
    This algorithm allows you to combine the outputs of multiple separation 
    algorithms, fusing them into a single output via clustering. It was first
    developed in [1]. When used with primitive separation algorithms, it becomes
    the PrimitiveClustering algorithm described in [1].

    References:

    [1] Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
        Diss. Northwestern University, 2019.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.

        num_sources (int): Number of sources to separate from signal.

        separators (list): List of instantiated separation algorithms that will be
          run on the input audio signal.

        weights (list, optional): Weight to give to each algorithm in the resultant
          feature vector. For example, `[3, 1]`, will repeat the features from the first
          algorithm 3 times and the second algorithm 1 time. Defaults to None - every
          algorithm gets a weight of 1.

        returns (list, optional): Which outputs of each algorithm to keep in the resultant
          feature vector. Defaults to None.

        num_cascades (int, optional): The output of each algorithm can be cascaded into 
          one another. The outputs of the first layer of algorithms will be refed to 
          each separation algorithm to create more features. Defaults to 1.

        extracted_feature (str, optional): Which feature to extract from each algorithm.
          Must be one of `['estimates', 'masks']`. `estimates` will reconstruct a soft
          mask using the output of the algorithm (useful if the algorithm is not a masking
          based separation algorithm). `masks` will use the data in the `result_masks`
          attribute of the separation algorithm. Defaults to 'masks'.

        clustering_type (str): One of 'KMeans', 'GaussianMixture', and 'MiniBatchKMeans'.
          The clustering approach to use on the features. Defaults to 'KMeans'.

        fit_clusterer (bool, optional): Whether or not to call fit on the clusterer.
          If False, then the clusterer should already be fit for this to work. Defaults
          to True.

        percentile (int, optional): Percentile of time-frequency points to consider by loudness. 
          Audio spectrograms are very high dimensional, and louder points tend to 
          matter more than quieter points. By setting the percentile high, one can more
          efficiently cluster an auditory scene by considering only points above
          that threshold. Defaults to 90 (which means the top 10 percentile of 
          time-frequency points will be used for clustering).

        beta (float, optional): When using KMeans, we use soft KMeans, which has an additional 
          parameter `beta`. `beta` controls how soft the assignments are. As beta 
          increases, the assignments become more binary (either 0 or 1). Defaults to 
          5.0, a value discovered through cross-validation.

        mask_type (str, optional): Masking approach to use. Passed up to MaskSeparationBase.

        mask_threshold (float, optional): Threshold for masking. Passed up to MaskSeparationBase.

        **kwargs (dict, optional): Additional keyword arguments that are passed to the clustering
          object (one of KMeans, GaussianMixture, or MiniBatchKMeans).
    
    Example:
        
        .. code-block:: python

            from nussl.separation import (
                primitive, 
                factorization, 
                composite, 
                SeparationException
            )

            separators = [
                primitive.FT2D(mix),
                factorization.RPCA(mix),
                primitive.Melodia(mix, voicing_tolerance=0.2),
                primitive.HPSS(mix),
            ]

            weights = [3, 3, 1, 1]
            returns = [[1], [1], [1], [0]]

            ensemble = composite.EnsembleClustering(
                mix, 2, separators, weights=weights, returns=returns)
            estimates = ensemble()
    """
    def __init__(self, input_audio_signal, num_sources, separators, weights=None, 
                 returns=None, num_cascades=1, extracted_feature='masks', 
                 clustering_type='KMeans', fit_clusterer=True, percentile=90,
                 beta=5.0, mask_type='soft', mask_threshold=0.5, **kwargs):
        super().__init__(
            input_audio_signal, num_sources, clustering_type=clustering_type, 
            percentile=percentile, fit_clusterer=fit_clusterer, beta=beta, 
            mask_type=mask_type, mask_threshold=mask_threshold, **kwargs)

        self.separators = separators
        self.num_cascades = num_cascades

        if isinstance(weights, list):
            if len(weights) != len(separators):
                raise SeparationException(
                    f"len(weights) must be the same as len(separators)!")
            self.weights = weights
        else:
            self.weights = [1 for _ in range(len(self.separators))]

        if isinstance(returns, list):
            if len(returns) != len(separators):
                raise SeparationException(
                    f"len(returns) must be the same as len(separators)!")
            self.returns = returns
        else:
            self.returns = None

        if extracted_feature not in ['masks', 'estimates']:
            raise SeparationException(
                f"extracted_feature must be one of ['masks', 'estimates']. "
                f"Got {extracted_feature}.")

        self.extracted_feature = extracted_feature

    def run_separators_on_mixture(self, mixture):
        estimates = []
        masks = []
        for i, separator in enumerate(self.separators):
            weight = self.weights[i]

            _estimates = separator(audio_signal=mixture)
            _names = [str(separator) for _ in _estimates]
            _masks = []

            if hasattr(separator, 'result_masks'):
                _masks = separator.result_masks

            if self.returns is not None:
                returns = self.returns[i]
                _estimates = [_estimates[j] for j in returns]
                if _masks:
                    _masks = [_masks[j] for j in returns]

            for _ in range(weight):
                estimates.extend(_estimates)
                masks.extend(_masks)

        return estimates, masks

    def extract_features(self):
        features = []
        current_signal = [self.audio_signal]

        for i in range(self.num_cascades):
            new_signals = []

            for _signal in current_signal:
                estimates, masks = self.run_separators_on_mixture(_signal)
                new_signals.extend(estimates)

                if self.extracted_feature == 'masks':
                    _features = self._extract_features_from_masks(masks)
                elif self.extracted_feature == 'estimates':
                    _features = self._extract_features_from_estimates(estimates)

                features.append(_features)
            
            current_signal = new_signals
        
        features = np.concatenate(features, axis=-1)

        return features

    def _extract_features_from_estimates(self, estimates):
        features = []
        mix_stft = np.abs(self.stft)
        for e in estimates:
            _stft = np.abs(e.stft())
            data = _stft / np.maximum(_stft, mix_stft + 1e-7)
            features.append(data)
        features = np.stack(features, axis=-1)
        return features

    @staticmethod
    def _extract_features_from_masks(masks):
        features = []
        for m in masks:
            mask_data = m.mask
            features.append(mask_data)
        features = np.stack(features, axis=-1)
        return features



import numpy as np
import norbert

from ..base import MaskSeparationBase, SeparationException


class WienerFilter(MaskSeparationBase):
    """
    Implements a multichannel Wiener filter that is computed by using some
    source estimates. When using the estimates produced by IdealRatioMask or 
    IdealBinaryMask, this is one of the upper baselines.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        estimates (list): List of audio signal objects that correspond to the estimates.
        iterations (int): Number of iterations for expectation-maximization in Wiener 
          filter.
        mask_type (str, optional): Mask type. Defaults to 'soft'.
        mask_threshold (float, optional): Threshold for masking binary. Defaults to 0.5.
        kwargs (dict): Additional keyword arguments to `norbert.wiener`.
    """

    def __init__(self, input_audio_signal, estimates, iterations=1, mask_type='soft',  
                 mask_threshold=.5, **kwargs):
        if not isinstance(estimates, list):
            raise SeparationException("estimates must be a list!")

        self.estimates = estimates
        self.iterations = iterations
        self.kwargs = kwargs

        super().__init__(
            input_audio_signal=input_audio_signal, 
            mask_type=mask_type,
            mask_threshold=mask_threshold)

    def run(self):
        source_magnitudes = np.stack([
            np.abs(e.stft()) for e in self.estimates], axis=-1)
        source_magnitudes = np.transpose(source_magnitudes, (1, 0, 2, 3))
        mix_stft = np.transpose(self.audio_signal.stft(), (1, 0, 2))

        enhanced = norbert.wiener(
            source_magnitudes, mix_stft, 
            iterations=self.iterations, **self.kwargs)
        _masks = np.abs(enhanced) / np.maximum(1e-7, np.abs(mix_stft[..., None]))
        _masks = np.transpose(_masks, (1, 0, 2, 3))

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)            
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)
        
        return self.result_masks



import numpy as np
import museval

from .evaluation_base import EvaluationBase


def _scale_bss_eval(references, estimate, idx, compute_sir_sar=True):
    """
    Helper for scale_bss_eval to avoid infinite recursion loop.
    """
    source = references[..., idx]
    source_energy = (source ** 2).sum()

    alpha = (
        source @ estimate / source_energy
    )

    e_true = source
    e_res = estimate - e_true

    signal = (e_true ** 2).sum()
    noise = (e_res ** 2).sum()

    snr = 10 * np.log10(signal / noise)

    e_true = source * alpha
    e_res = estimate - e_true

    signal = (e_true ** 2).sum()
    noise = (e_res ** 2).sum()

    si_sdr = 10 * np.log10(signal / noise)

    srr = -10 * np.log10((1 - (1/alpha)) ** 2)
    sd_sdr = snr + 10 * np.log10(alpha ** 2)

    si_sir = np.nan
    si_sar = np.nan

    if compute_sir_sar:
        references_projection = references.T @ references

        references_onto_residual = np.dot(references.transpose(), e_res)
        b = np.linalg.solve(references_projection, references_onto_residual)

        e_interf = np.dot(references, b)
        e_artif = e_res - e_interf

        si_sir = 10 * np.log10(signal / (e_interf ** 2).sum())
        si_sar = 10 * np.log10(signal / (e_artif ** 2).sum())

    return si_sdr, si_sir, si_sar, sd_sdr, snr, srr


def scale_bss_eval(references, estimate, mixture, idx, 
                   compute_sir_sar=True):
    """
    Computes metrics for references[idx] relative to the
    chosen estimates. This only works for mono audio. Each
    channel should be done independently when calling this
    function. Lovingly borrowed from Gordon Wichern and 
    Jonathan Le Roux at Mitsubishi Electric Research Labs.

    This returns 9 numbers (in this order):

    - SI-SDR: Scale-invariant source-to-distortion ratio. Higher is better.
    - SI-SIR: Scale-invariant source-to-interference ratio. Higher is better.
    - SI-SAR: Scale-invariant source-to-artifact ratio. Higher is better.
    - SD-SDR: Scale-dependent source-to-distortion ratio. Higher is better.
    - SNR: Signal-to-noise ratio. Higher is better.
    - SRR: The source-to-rescaled-source ratio. This corresponds to 
      a term that punishes the estimate if its scale is off relative
      to the reference. This is an unnumbered equation in [1], but
      is the term on page 2, second column, second to last line:
      ||s - alpha*s||**2. s here is factored out. Higher is better.
    - SI-SDRi: Improvement in SI-SDR over using the mixture as the estimate.
    - SD-SDRi: Improvement in SD-SDR over using the mixture as the estimate.
    - SNRi: Improvement in SNR over using the mixture as the estimate.

    References:

    [1] Le Roux, J., Wisdom, S., Erdogan, H., & Hershey, J. R. 
        (2019, May). SDR–half-baked or well done?. In ICASSP 2019-2019 IEEE 
        International Conference on Acoustics, Speech and Signal 
        Processing (ICASSP) (pp. 626-630). IEEE.
    
    Args:
        references (np.ndarray): object containing the
          references data. Of shape (n_samples, n_sources).
         
        estimate (np.ndarray): object containing the
          estimate data. Of shape (n_samples, 1).

        mixture (np.ndarray): objct containingthe
          mixture data. Of shape (n_samples, 1).

        idx (int): Which reference to compute metrics against.

        compute_sir_sar (bool, optional): Whether or not to compute SIR/SAR
          metrics, which can be computationally expensive and may not be
          relevant for your evaluation. Defaults to True

    Returns:
        tuple: SI-SDR, SI-SIR, SI-SAR, SD-SDR, SNR, SRR, SI-SDRi, SD-SDRi, SNRi
    """    
    si_sdr, si_sir, si_sar, sd_sdr, snr, srr = _scale_bss_eval(
        references, estimate, idx, compute_sir_sar=compute_sir_sar)
    mix_metrics = _scale_bss_eval(
        references, mixture, idx, compute_sir_sar=False)

    mix_si_sdr = mix_metrics[0]
    mix_sd_sdr = mix_metrics[3]
    mix_snr = mix_metrics[4]
    si_sdri = si_sdr - mix_metrics[0]
    sd_sdri = sd_sdr - mix_metrics[3]
    snri = snr - mix_metrics[4]

    return (
      si_sdr, si_sir, si_sar, sd_sdr, snr, srr, si_sdri, sd_sdri, snri, 
      mix_si_sdr, mix_sd_sdr, mix_snr
    )


class BSSEvaluationBase(EvaluationBase):
    """
    Base class for all evaluation classes that are based on BSSEval metrics. This 
    contains some useful verification functions, preprocessing functions that are
    used in many separation-based evaluation. Specific evaluation metrics are 
    thin wrappers around this base class, basically only implementing the
    ``self.evaluate_helper`` function.
    
    Both ``true_sources_list`` and ``estimated_sources_list`` get validated 
    using the private method :func:`_verify_input_list`. If your evaluation 
    needs to verify that input is set correctly (recommended) overwrite that method 
    to add checking.
    
    Args:
        true_sources_list (list): List of objects that contain one ground truth source per object.
          In some instances (such as the :class:`BSSEval` objects) this list is filled with
          :class:`AudioSignals` but in other cases it is populated with
          :class:`MaskBase` -derived objects (i.e., either a :class:`BinaryMask` or
          :class:`SoftMask` object).
        estimated_sources_list (list): List of objects that contain source estimations from a source
          separation algorithm. List should be populated with the same type of objects and in the
          same order as :param:`true_sources_list`.
        source_labels (list): List of strings that are labels for each source to be used as keys for
          the scores. Default value is `None` and in that case labels use the file_name attribute.
          If that is also `None`, then the source labels are `Source 0`, `Source 1`, etc.
        compute_permutation (bool): Whether or not to evaluate in a permutation-invariant 
          fashion, where the estimates are permuted to match the true sources. Only the 
          best permutation according to ``best_permutation_key`` is returned to the 
          scores dict. Defaults to False.
        best_permutation_key (str): Which metric to use to decide which permutation of 
          the sources was best.
        **kwargs (dict): Any additional arguments are passed on to evaluate_helper.
    """
    # TODO: Populate score in evaluation_helper() using self.keys.
    keys = ['SDR', 'ISR', 'SIR', 'SAR']
    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None,
                 compute_permutation=False, best_permutation_key="SDR", **kwargs):
        super().__init__(true_sources_list, estimated_sources_list, source_labels=source_labels,
                         compute_permutation=compute_permutation,
                         best_permutation_key=best_permutation_key,
                         **kwargs)

    def preprocess(self):
        """
        Implements preprocess by stacking the audio_data inside each AudioSignal
        object in both self.true_sources_list and self.estimated_sources_list.
        
        Returns:
            tuple: Tuple containing reference and estimate arrays.
        """
        references = np.stack(
            [x.audio_data for x in self.true_sources_list],
            axis=-1
        )
        estimates = np.stack(
            [x.audio_data for x in self.estimated_sources_list],
            axis=-1
        )
        return references.transpose((1, 0, 2)), estimates.transpose((1, 0, 2))


class BSSEvalV4(BSSEvaluationBase):
    def evaluate_helper(self, references, estimates, **kwargs):
        """
        Implements evaluation using museval.metrics.bss_eval
        """
        # museval expects shape=(nsrc, nsampl, nchan)
        # we have (nsampl, nchan, nsrc)
        # so let's massage the data so it matches before feeding it in

        references = np.transpose(references, (2, 0, 1))
        estimates = np.transpose(estimates, (2, 0, 1))

        sdr, isr, sir, sar, _ = museval.metrics.bss_eval(
            references, estimates, compute_permutation=False, **kwargs)

        scores = []
        for j in range(references.shape[0]):
            score = {
                'SDR': sdr[j].tolist(), 
                'ISR': isr[j].tolist(), 
                'SIR': sir[j].tolist(), 
                'SAR': sar[j].tolist(),
            }
            scores.append(score)
        return scores


class BSSEvalScale(BSSEvaluationBase):
    # TODO: Populate score in evaluation_helper() using self.keys.
    keys = [
        'SI-SDR', 'SI-SIR', 'SI-SAR',
        'SD-SDR', 'SNR', 'SRR',
        'SI-SDRi', 'SD-SDRi', 'SNRi',
        'MIX-SI-SDR', 'MIX-SD-SDR', 'MIX-SNR',
    ]
    def __init__(self, true_sources_list, estimated_sources_list, source_labels=None,
                 compute_permutation=False, best_permutation_key="SI-SDR", **kwargs):
        super().__init__(true_sources_list, estimated_sources_list, source_labels=source_labels,
                         compute_permutation=compute_permutation,
                         best_permutation_key=best_permutation_key,
                         **kwargs)

    def preprocess(self):
        """
        Scale invariant metrics expects zero-mean centered references and sources.
        """
        references, estimates = super().preprocess()

        mixture = references.sum(axis=-1)
        mixture -= mixture.mean(axis=0)

        self.mixture = mixture
        references -= references.mean(axis=0)
        estimates -= estimates.mean(axis=0)

        return references, estimates

    def evaluate_helper(self, references, estimates, compute_sir_sar=True):
        """
        Implements evaluation using new BSSEval metrics [1]. This computes every
        metric described in [1], including:

        - SI-SDR: Scale-invariant source-to-distortion ratio. Higher is better.
        - SI-SIR: Scale-invariant source-to-interference ratio. Higher is better.
        - SI-SAR: Scale-invariant source-to-artifact ratio. Higher is better.
        - SD-SDR: Scale-dependent source-to-distortion ratio. Higher is better.
        - SNR: Signal-to-noise ratio. Higher is better.
        - SRR: The source-to-rescaled-source ratio. This corresponds to 
          a term that punishes the estimate if its scale is off relative
          to the reference. This is an unnumbered equation in [1], but
          is the term on page 2, second column, second to last line:
          ||s - alpha*s||**2. s is factored out. Higher is better.
        - SI-SDRi: Improvement in SI-SDR over using the mixture as the estimate. Higher 
          is better.
        - SD-SDRi: Improvement in SD-SDR over using the mixture as the estimate. Higher
          is better.
        - SNRi: Improvement in SNR over using the mixture as the estimate. Higher is
          better.

        Note:

        If `compute_sir_sar = False`, then you'll get `np.nan` for SI-SIR and 
        SI-SAR!

        References:

        [1] Le Roux, J., Wisdom, S., Erdogan, H., & Hershey, J. R. 
        (2019, May). SDR–half-baked or well done?. In ICASSP 2019-2019 IEEE 
        International Conference on Acoustics, Speech and Signal 
        Processing (ICASSP) (pp. 626-630). IEEE.
        """

        sisdr, sisir, sisar, sdsdr, snr, srr = [], [], [], [], [], []
        sisdri, sdsdri, snri, mix_sisdr, mix_sdsdr, mix_snr = [], [], [], [], [], []
        for j in range(references.shape[-1]):
            cSISDR, cSISIR, cSISAR, cSDSDR, cSNR, cSRR = [], [], [], [], [], []
            cSISDRi, cSDSDRi, cSNRi, cMIXSISDR, cMIXSDSDR, cMIXSNR = [], [], [], [], [], []
            for ch in range(references.shape[-2]):
                output = scale_bss_eval(
                        references[..., ch, :], estimates[..., ch, j], 
                        self.mixture[..., ch], j, compute_sir_sar=compute_sir_sar
                    )
                _SISDR, _SISIR, _SISAR, _SDSDR, _SNR, _SRR, = output[:6]
                _SISDRi, _SDSDRi, _SNRi, _MIXSISDR, _MIXSDSDR, _MIXSNR = output[6:]
            
                cSISDR.append(_SISDR)
                cSISIR.append(_SISIR)
                cSISAR.append(_SISAR)
                cSDSDR.append(_SDSDR)
                cSNR.append(_SNR)
                cSRR.append(_SRR)
                cSISDRi.append(_SISDRi)
                cSDSDRi.append(_SDSDRi)
                cSNRi.append(_SNRi)
                cMIXSISDR.append(_MIXSISDR) 
                cMIXSDSDR.append(_MIXSDSDR)
                cMIXSNR.append(_MIXSNR)

            sisdr.append(cSISDR)
            sisir.append(cSISIR)
            sisar.append(cSISAR)

            sdsdr.append(cSDSDR)
            snr.append(cSNR)
            srr.append(cSRR)

            sisdri.append(cSISDRi)
            sdsdri.append(cSDSDRi)
            snri.append(cSNRi)

            mix_sisdr.append(cMIXSISDR)
            mix_sdsdr.append(cMIXSDSDR)
            mix_snr.append(cMIXSNR)

        scores = []
        for j in range(references.shape[-1]):
            score = {
                'SI-SDR': sisdr[j], 'SI-SIR': sisir[j], 'SI-SAR': sisar[j],
                'SD-SDR': sdsdr[j], 'SNR': snr[j], 'SRR': srr[j], 
                'SI-SDRi': sisdri[j], 'SD-SDRi': sdsdri[j], 'SNRi': snri[j],
                'MIX-SI-SDR': mix_sisdr[j], 'MIX-SD-SDR': mix_sdsdr[j],
                'MIX-SNR': mix_snr[j],
            }
            scores.append(score)
        return scores



import numpy as np
from scipy.ndimage.filters import maximum_filter, minimum_filter, uniform_filter

from .. import MaskSeparationBase, SeparationException
from ..benchmark import HighLowPassFilter


class FT2D(MaskSeparationBase):
    """
    This separation method is based on using 2DFT image processing for source
    separation [1].

    The algorithm has five main steps:

    1. Take the 2DFT of Magnitude STFT.
    2. Identify peaks in the 2DFT (these correspond to repeating patterns).
    3. Mask everything but the peaks and invert the masked 2DFT back to a magnitude STFT.
    4. Take the residual peakless 2DFT and invert that to a magnitude STFT.
    5. Compare the two magnitude STFTs to construct masks for the foreground and
       the background.

    The algorithm runs on each channel independently. The masks can be constructed
    in two ways: either using the background (peaky) 2DFT as the reference for making 
    masks, or using the foreground (peakless) 2DFT as the reference for making masks.
    This behavior can be toggled via `use_bg_2dft=[True, False]`. Using the background
    biases the algorithm towards separating repeating patterns whereas using
    the foreground biases it towards preserving micromodulation in the foreground.

    There are two ways to identify peaks: either using the `original` method, as
    laid out in the paper which identifies peaks in a binary way as being above
    some threshold, or by using the `local_std` way, which was developed and investigated
    in Chapter 2 in [2]. This method identifies peaks in a soft way and can thus
    be used to construct soft masks. It generally performs better.

    The main hyperparameter to consider is the `neighborhood_size`, which determines
    how big of a filter to apply when looking for peaks. It is two dimensional, but
    generally keeping 1D horizontal filters is the way to go.

    References:

        [1] Seetharaman, Prem, Fatemeh Pishdadian, and Bryan Pardo.
            "Music/Voice Separation Using the 2D Fourier Transform." 
            2017 IEEE Workshop on Applications of Signal Processing to 
            Audio and Acoustics (WASPAA). IEEE, 2017.

        [2] Seetharaman, Prem. Bootstrapping the Learning Process for Computer Audition. 
            Diss. Northwestern University, 2019.

    Args:
        input_audio_signal (AudioSignal): Signal to separate.

        neighborhood_size (tuple, optional): 2-tuple of ints telling the filter size to look for
          peaks in frequency and time: `(f, t)`. Defaults to (1, 25).

        high_pass_cutoff (float, optional): Time-frequency bins below this cutoff will be
          assigned to the background. Defaults to 100.0.

        quadrants_to_keep (tuple, optional): Each quadrant of the 2DFT can be filtered out
          when separating. Can be used for separating out upward spectro-temporal patterns
          (via `(0, 2)` from downward ones `(1, 3)`. Defaults to (0,1,2,3).

        filter_approach (str, optional): One of 'original' or 'local_std'. Which 
          filtering approach to apply to identify peaks. Defaults to 'local_std'.

        use_bg_2dft (bool, optional): Whether to use the background or foreground 2DFT
          as the reference for constructing masks. Defaults to True.

        mask_type (str, optional): Mask type. Defaults to 'soft'.

        mask_threshold (float, optional): Masking threshold. Defaults to 0.5.

    """

    def __init__(self, input_audio_signal, neighborhood_size=(1, 25), high_pass_cutoff=100.0,
                 quadrants_to_keep=(0, 1, 2, 3), filter_approach='local_std', use_bg_2dft=True,
                 mask_type='soft', mask_threshold=0.5):

        super().__init__(
            input_audio_signal=input_audio_signal,
            mask_type=mask_type,
            mask_threshold=mask_threshold
        )

        self.neighborhood_size = neighborhood_size
        self.result_masks = None
        self.quadrants_to_keep = quadrants_to_keep
        self.use_bg_2dft = use_bg_2dft
        self.high_pass_cutoff = high_pass_cutoff
        self.bg_ft2d = None
        self.fg_ft2d = None

        if filter_approach == 'original':
            self.filter_func = self.filter_local_maxima
        elif filter_approach == 'local_std':
            self.filter_func = self.filter_local_maxima_with_std
        else:
            allowed_filter_approaches = ['original', 'local_std']
            raise SeparationException(
                f'filter approach must be one of {allowed_filter_approaches}')

    def run(self):
        high_low = HighLowPassFilter(self.audio_signal, self.high_pass_cutoff)
        high_pass_masks = high_low.run()

        background_masks = []
        foreground_masks = []

        for ch in range(self.audio_signal.num_channels):
            background_mask, foreground_mask = self.compute_ft2d_mask(
                self.ft2d, ch
            )
            background_masks.append(background_mask)
            foreground_masks.append(foreground_mask)

        background_masks = np.stack(background_masks, axis=-1)
        foreground_masks = np.stack(foreground_masks, axis=-1)

        _masks = np.stack([background_masks, foreground_masks], axis=-1)

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)

            if i == 0:
                mask_data = np.maximum(mask_data, high_pass_masks[i].mask)
            elif i == 1:
                mask_data = np.minimum(mask_data, high_pass_masks[i].mask)

            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks

    def _preprocess_audio_signal(self):
        super()._preprocess_audio_signal()

        self.ft2d = np.stack([
            np.fft.fft2(np.abs(self.stft[:, :, i]))
            for i in range(self.audio_signal.num_channels)],
            axis=-1
        )

    @staticmethod
    def filter_quadrants(data, quadrants_to_keep):
        """
        Filters the quadrants of the incoming 2DFT, deleting the ones not in 
        `quadrants_to_keep`. Those are zero'd out. This can be used to separate
        upward moving from downward moving spectro-temporal patterns.
        
        .. code-block:: none
        
            1: shape[0] // 2:, :shape[1] // 2
            2: :shape[0] // 2, :shape[1] // 2
            3: :shape[0] // 2, shape[1] // 2:
            4: shape[0] // 2:, shape[1] // 2:
        
        Args:
            data ([type]): [description]
            quadrants_to_keep ([type]): [description]
        
        Returns:
            [type]: [description]
        """

        shape = data.shape
        for quadrant in range(4):
            if quadrant not in quadrants_to_keep:
                if quadrant == 0:
                    data[shape[0] // 2:, :shape[1] // 2] = 0
                elif quadrant == 1:
                    data[:shape[0] // 2, :shape[1] // 2] = 0
                elif quadrant == 2:
                    data[:shape[0] // 2, shape[1] // 2:] = 0
                elif quadrant == 3:
                    data[shape[0] // 2:, shape[1] // 2:] = 0
        return data

    def compute_ft2d_mask(self, ft2d, ch):
        bg_ft2d, fg_ft2d = self.filter_func(ft2d[:, :, ch])

        self.bg_ft2d = self.filter_quadrants(bg_ft2d, self.quadrants_to_keep)
        self.fg_ft2d = self.filter_quadrants(fg_ft2d, self.quadrants_to_keep)

        _stft = np.abs(self.stft)[:, :, ch] + 1e-7
        _stft = _stft

        if self.use_bg_2dft:
            ft2d_used = self.bg_ft2d
        else:
            ft2d_used = self.fg_ft2d

        est_stft = np.minimum(np.abs(np.fft.ifft2(ft2d_used)), _stft)
        est_mask = est_stft / _stft
        est_mask /= (est_mask + 1e-7).max()

        if self.use_bg_2dft:
            bg_mask = est_mask
            fg_mask = 1 - bg_mask
        else:
            fg_mask = est_mask
            bg_mask = 1 - fg_mask

        return bg_mask, fg_mask

    def filter_local_maxima_with_std(self, ft2d):
        data = np.abs(np.fft.fftshift(ft2d))
        data /= (np.max(data) + 1e-7)

        data_max = maximum_filter(data, self.neighborhood_size)
        data_mean = uniform_filter(data, self.neighborhood_size)
        data_mean_sq = uniform_filter(data ** 2, self.neighborhood_size)
        data_std = np.sqrt(data_mean_sq - data_mean ** 2) + 1e-7

        maxima = ((data_max - data_mean) / data_std)
        fraction_of_local_max = (data == data_max)
        maxima *= fraction_of_local_max
        maxima = maxima.astype(float)
        maxima /= (np.max(maxima) + 1e-7)

        maxima = np.maximum(maxima, np.fliplr(maxima), np.flipud(maxima))
        maxima = np.fft.ifftshift(maxima)

        background_ft2d = np.multiply(maxima, ft2d)
        foreground_ft2d = np.multiply(1 - maxima, ft2d)

        return background_ft2d, foreground_ft2d

    def filter_local_maxima(self, ft2d):
        data = np.abs(np.fft.fftshift(ft2d))
        data /= (np.max(data) + 1e-7)
        threshold = np.std(data)

        data_max = maximum_filter(data, self.neighborhood_size)
        maxima = (data == data_max)
        data_min = minimum_filter(data, self.neighborhood_size)
        diff = ((data_max - data_min) > threshold)
        maxima[diff == 0] = 0
        maxima = np.maximum(maxima, np.fliplr(maxima), np.flipud(maxima))
        maxima = np.fft.ifftshift(maxima)

        background_ft2d = np.multiply(maxima, ft2d)
        foreground_ft2d = np.multiply(1 - maxima, ft2d)

        return background_ft2d, foreground_ft2d



"""
Base class for separation algorithms that make masks. Most algorithms in 
nussl are derived from MaskSeparationBase. 
"""

from ...core import masks
from . import SeparationBase
from .separation_base import SeparationException


class MaskSeparationBase(SeparationBase):
    """
    Base class for separation algorithms that create a mask (binary or soft) to do 
    their separation. Most algorithms in nussl are derived from 
    :class:`MaskSeparationBase`.
    
    Although this class will do nothing if you instantiate and run it by itself,
    algorithms that are derived from this class are expected to return a list of 
    :class:`separation.masks.mask_base.MaskBase` -derived objects  
    (i.e., either a :class:`separation.masks.binary_mask.BinaryMask` or 
    :class:`separation.masks.soft_mask.SoftMask` object) by their :func:`run()` 
    method. Being a subclass of :class:`MaskSeparationBase` is an implicit contract 
    assuring this.  Returning a :class:`separation.masks.mask_base.MaskBase`-derived 
    object standardizes algorithm return types for 
    :class:`evaluation.evaluation_base.EvaluationBase`-derived objects. 
    
    Args:
        input_audio_signal: (:class:`audio_signal.AudioSignal`) An 
          :class:`audio_signal.AudioSignal` object containing the mixture to be 
          separated.
        mask_type: (str, BinaryMask, or SoftMask) Indicates whether to make 
          binary or soft masks. See :attr:`mask_type` property for details.
        mask_threshold: (float) Value between [0.0, 1.0] to convert a soft mask 
          to a binary mask. See :attr:`mask_threshold` property for details.
    """
    
    MASKS = {
        'binary': masks.BinaryMask,
        'soft': masks.SoftMask
    }

    def __init__(self, input_audio_signal, mask_type='soft', mask_threshold=0.5):
        super().__init__(input_audio_signal=input_audio_signal)

        self.mask_type = mask_type
        self.mask_threshold = mask_threshold
        self.result_masks = []

        self.metadata.update({
            'mask_type': mask_type,
            'mask_threshold': 'N/A' if self.mask_type == masks.SoftMask else mask_threshold
        })

    @property
    def mask_type(self):
        """        
        This property indicates what type of mask the derived algorithm will create 
        and be returned by :func:`run()`. Options are either 'soft' or 'binary'. 
        :attr:`mask_type` is usually set when initializing a 
        :class:`MaskSeparationBase`-derived class and defaults to 'soft..
        
        This property, though stored as a string, can be set in two ways when 
        initializing:
        
        * First, it is possible to set this property with a string. Only ``'soft'`` 
          and ``'binary'`` are accepted (case insensitive), every other value will 
          raise an error. When initializing with a string, two helper 
          attributes are provided: :attr:`BINARY_MASK` and :attr:`SOFT_MASK`.
        
          It is **HIGHLY** encouraged to use these, as the API may change and code 
          that uses bare strings (e.g. ``mask_type = 'soft'`` or 
          ``mask_type = 'binary'``) for assignment might not be future-proof. 
          :attr:`BINARY_MASK`` and :attr:`SOFT_MASK` are safe aliases in case these 
          underlying types change.
        
        * The second way to set this property is by using a class prototype of 
          either the :class:`separation.masks.binary_mask.BinaryMask` or 
          :class:`separation.masks.soft_mask.SoftMask` class prototype. This is 
          probably the most stable way to set this, and it's fairly succinct. 
          For example, ``mask_type = nussl.BinaryMask`` or 
          ``mask_type = nussl.SoftMask`` are both perfectly valid.
        
        Though uncommon, this can be set outside of :func:`__init__()` 
        
        Examples of both methods are shown below.
        
        Returns:
            mask_type (str): Either ``'soft'`` or ``'binary'``. 
            
        Raises:
            ValueError if set invalidly.
            
        Example:
            
        .. code-block:: python
            :linenos:
    
            import nussl
            mixture_signal = nussl.AudioSignal()
                
            # Two options for determining mask upon init...
             
            # Option 1: Init with a string (BINARY_MASK is a string 'constant')
            repet_sim = nussl.RepetSim(mixture_signal, mask_type=nussl.MaskSeparationBase.BINARY_MASK)
             
            # Option 2: Init with a class type
            ola = nussl.OverlapAdd(mixture_signal, mask_type=nussl.SoftMask)
             
            # It's also possible to change these values after init by changing the `mask_type` property...
            repet_sim.mask_type = nussl.MaskSeparationBase.SOFT_MASK  # using a string
            ola.mask_type = nussl.BinaryMask  # or using a class type
            
        """
        return self._mask_type

    @mask_type.setter
    def mask_type(self, value):
        error = ValueError(
            f"Invalid mask type! Got {value} but valid masks are:"
            f" [{', '.join(list(self.MASKS.keys()))}]!"
        )

        if value is None:
            raise error

        if isinstance(value, str):
            value = value.lower()
            if value in self.MASKS:
                self._mask_type = self.MASKS[value]
            else:
                raise error
        elif isinstance(value, masks.BinaryMask):
            self._mask_type = self.MASKS['binary']
        elif isinstance(value, masks.SoftMask):
            self._mask_type = self.MASKS['soft']
        else:
            raise error

    @property
    def mask_threshold(self):
        """        
        Threshold of determining True/False if :attr:`mask_type` is 
        :attr:`BINARY_MASK`. Some algorithms will first make a soft mask and then 
        convert that to a binary mask using this threshold parameter. All 
        values of the soft mask are between ``[0.0, 1.0]`` and as such 
        :func:`mask_threshold` is expected to be a float between 
        ``[0.0, 1.0]``.
        
        Returns:
            mask_threshold (float): Value between ``[0.0, 1.0]`` that indicates 
              the True/False cutoff when converting a soft mask to binary mask.
                                
        Raises:
            ValueError if not a float or if set outside ``[0.0, 1.0]``.

        """
        return self._mask_threshold

    @mask_threshold.setter
    def mask_threshold(self, value):
        if not isinstance(value, float) or not (0.0 < value < 1.0):
            raise ValueError('Mask threshold must be a float between [0.0, 1.0]!')

        self._mask_threshold = value

    def zeros_mask(self, shape):
        """
        Creates a new zeros mask with this object's type.

        Args:
            shape (tuple): tuple with shape of mask

        Returns:
            A subclass of `MaskBase` containing 0s.
        """
        return self.mask_type.zeros(shape)

    def ones_mask(self, shape):
        """
        Creates a new ones mask with this object's type.

        Args:
            shape (tuple): tuple with shape of mask

        Returns:
            A subclass of `MaskBase` containing 1s.
        """
        return self.mask_type.ones(shape)

    def _preprocess_audio_signal(self):
        """
        Masking based separation algorithm always need an STFT to work with. 
        So here, the STFT of the AudioSignal object belonging to this separation
        algorithm is taken. It also resets the `self.result_masks` object to
        an empty list - new audio signal means new masks.

        This gets called when the `self.audio_signal` is set.
        """
        self.stft = self.audio_signal.stft()
        self.result_masks = []

    def run(self):
        """Runs mask-based separation algorithm. Base class: Do not call directly!

        Raises:
            NotImplementedError: Cannot call base class!
        """
        raise NotImplementedError('Cannot call base class!')

    def make_audio_signals(self):
        """
        Makes :class:`audio_signal.AudioSignal` objects after mask-based
        separation algorithm is run. This looks in ``self.result_masks``
        which must be filled by ``run`` in the algorithm that
        subclasses this. It applies each mask to the mixture audio 
        signal and returns a list of the estimates, which are each
        AudioSignal objects.

        Returns:
            list: List of AudioSignal objects corresponding to the 
              separated estimates.
        """
        if not self.result_masks:
            raise SeparationException(
                "self.result_masks is empty! Did you call self.run()?")
        
        estimates = []
        for mask in self.result_masks:
            if not isinstance(mask, self.mask_type):
                raise SeparationException(
                    f"Expected {self.mask_type} but got {type(mask)} "
                    f"in self.result_masks!"
                )
            estimate = self.audio_signal.apply_mask(mask, overwrite=False)
            estimate.istft()
            estimates.append(estimate)
        return estimates



"""
Deep unfolding is a type of architecture where an optimization
process like clustering, non-negative matrix factorization and
other EM style algorithms (anything with update functions) are
unfolded as layers in a neural network. In practice this results
in having the operations available to do on torch Tensors. This
submodule collects implementations that allow one to unfold different
optimization processes in a neural network.
"""

from .gaussian_mixture import GaussianMixtureTorch



"""
Machine Learning
================

SeparationModel
---------------

.. autoclass:: nussl.ml.SeparationModel
    :members:
    :autosummary:

Building blocks for SeparationModel
-----------------------------------

.. automodule:: nussl.ml.modules
    :members:
    :autosummary:

.. automodule:: nussl.ml.cluster
    :members:
    :autosummary:

Helpers for common deep networks
--------------------------------

.. automodule:: nussl.ml.networks.builders
    :members:
    :autosummary:

Confidence measures
-------------------

.. automodule:: nussl.ml.confidence
    :members:
    :autosummary:

Training
--------

.. automodule:: nussl.ml.train
    :members:
    :autosummary:
"""

from .networks import SeparationModel
from .networks import modules

from . import train
from . import unfold
from . import cluster
from . import confidence

from sklearn.decomposition import NMF

def register_module(module):
    """
    Your custom modules can be registered with nussl's SeparationModel
    via this function. For example, if you have some module `ExampleModule`,
    you can register it as follows:

    .. code-block:: python

        class ExampleModule(nn.Module):
        def forward(self, data):
            data = data * 2
            return data

        ml.register_module(ExampleModule)

    You can now use the name `ExampleModule` in the config for a SeparationModel.
    
    Args:
        module ([type]): [description]
    """
    setattr(modules, module.__name__, module)


def register_loss(custom_loss):
    """
    Your custom loss functions can be registered with nussl's Closure
    via this function. For example, if you have some loss `ExampleLoss`,
    you can register it as follows:

    .. code-block:: python

        class ExampleLoss(nn.Module):
            DEFAULT_KEYS = {'key1': 'arg1', 'key2': 'arg2'}
            def forward(self, arg1, arg2):
                # do something
                return 0

        ml.register_loss(ExampleLoss)

    You can now use the name `ExampleLoss` in a loss dictionary that is passed
    to a closure.

    
    Args:
        custom_loss ([type]): [description]
    """
    setattr(train.loss, custom_loss.__name__, custom_loss)



import torch

from ..base import ClusteringSeparationBase, DeepMixin, SeparationException


class DeepClustering(DeepMixin, ClusteringSeparationBase):
    """
    Clusters the embedding produced by a deep model for every time-frequency point.
    This is the deep clustering source separation approach. It is flexible with
    the number of sources. It expects that the model outputs a dictionary where one
    of the keys is 'embedding'. This uses the `DeepMixin` class to load the model
    and set the audio signal's parameters to be appropriate for the model.
    
    Args:
        input_audio_signal: (AudioSignal`) An AudioSignal object containing the 
          mixture to be separated.
        num_sources (int): Number of sources to cluster the features of and separate
          the mixture.
        model_path (str, optional): Path to the model that will be used. Can be None, 
          so that you can initialize a class and load the model later.  
          Defaults to None.
        device (str, optional): Device to put the model on. Defaults to 'cpu'.
        extra_data (dict, optional): Any extra data that is to be passed at runtime
          to the SeparationModel.
        **kwargs (dict): Keyword arguments for ClusteringSeparationBase and the 
          clustering object used for clustering (one of KMeans, GaussianMixture,
          MiniBatchKmeans).
    
    Raises:
        SeparationException: If 'embedding' isn't in the output of the model.
    """
    def __init__(self, input_audio_signal, num_sources, model_path=None,
                 device='cpu', extra_data=None, **kwargs):
        super().__init__(input_audio_signal, num_sources, **kwargs)
        if model_path is not None:
            self.load_model(model_path, device=device)
        # audio channel dimension in a dpcl model
        self.channel_dim = -1
        self.extra_data = extra_data

    def forward(self):
        return self.extract_features()

    def extract_features(self):
        input_data = self._get_input_data_for_model(self.extra_data)
        with torch.no_grad():
            output = self.model(input_data)
            if 'embedding' not in output:
                raise SeparationException(
                    "This model is not a deep clustering model! "
                    "Did not find 'embedding' key in output dictionary.")
            embedding = output['embedding']
            # swap back batch and sample dims
            if self.metadata['num_channels'] == 1:
                embedding = embedding.transpose(0, -2)
            embedding = embedding.squeeze(0).transpose(0, 1)
        self._preprocess_audio_signal()
        return embedding.cpu().data.numpy()



import numpy as np
import scipy.fftpack as scifft

from .. import MaskSeparationBase, SeparationException
from ..benchmark import HighLowPassFilter
from ...core import constants


class Repet(MaskSeparationBase):
    """Implements the original REpeating Pattern Extraction Technique algorithm 
    using the beat spectrum.

    REPET is a simple method for separating a repeating background from a 
    non-repeating foreground in an audio mixture. It assumes a single repeating 
    period over the whole signal duration, and finds that period based on finding 
    a peak in the beat spectrum. The period can also be provided exactly, or you
    can give ``Repet`` a guess of the min and max period. Once it has a period, 
    it "overlays" spectrogram sections of length ``period`` to create a median 
    model (the background).
    
    References:

    [1] Rafii, Zafar, and Bryan Pardo.
        "Repeating pattern extraction technique (REPET): A simple method for 
        music/voice separation." IEEE transactions on audio, speech, 
        and language processing 21.1 (2012): 73-84.

    Args:
        input_audio_signal (AudioSignal): Signal to separate.

        min_period (float, optional): minimum time to look for repeating period in 
          terms of seconds.
        
        max_period (float, optional): maximum time to look for repeating period in 
          terms of seconds.
        
        period (float, optional): exact time that the repeating period is 
          (in seconds).
        
        high_pass_cutoff (float, optional): value (in Hz) for the high pass 
          cutoff filter.
                        
        mask_type (str, optional): Mask type. Defaults to 'soft'.

        mask_threshold (float, optional): Masking threshold. Defaults to 0.5.

    """

    def __init__(self, input_audio_signal, min_period=None, max_period=None,
                 period=None, high_pass_cutoff=100.0, mask_type='soft',
                 mask_threshold=0.5):

        super().__init__(
            input_audio_signal=input_audio_signal,
            mask_type=mask_type,
            mask_threshold=mask_threshold)

        # Check input parameters
        if (min_period or max_period) and period:
            raise SeparationException(
                'Cannot set both period and (min_period or max_period)!')

        self._is_period_converted_to_hops = False
        self.period = period

        if self.period is None:
            self.min_period = min_period if min_period else .8
            max_period = max_period if max_period else 8
            self.max_period = min(max_period, self.audio_signal.signal_duration / 3)
        else:
            self.period = self._update_period(period)
            self._is_period_converted_to_hops = True

        self.high_pass_cutoff = high_pass_cutoff
        self.magnitude_spectrogram = None
        self.repeating_period = None
        self.beat_spectrum = None

    def run(self):
        high_low = HighLowPassFilter(self.audio_signal, self.high_pass_cutoff)
        high_pass_masks = high_low.run()

        self.magnitude_spectrogram = np.abs(self.stft)

        background_masks = []
        foreground_masks = []

        self.repeating_period = self._calculate_repeating_period()

        for ch in range(self.audio_signal.num_channels):
            background_mask = self._compute_repeating_mask(
                self.magnitude_spectrogram[..., ch])
            foreground_mask = 1 - background_mask

            background_masks.append(background_mask)
            foreground_masks.append(foreground_mask)

        background_masks = np.stack(background_masks, axis=-1)
        foreground_masks = np.stack(foreground_masks, axis=-1)

        _masks = np.stack([background_masks, foreground_masks], axis=-1)

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)

            if i == 0:
                mask_data = np.maximum(mask_data, high_pass_masks[i].mask)
            elif i == 1:
                mask_data = np.minimum(mask_data, high_pass_masks[i].mask)

            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks

    def get_beat_spectrum(self):
        """
        Calculates and returns the beat spectrum for the audio signal associated 
        with this object

        Returns:
            beat_spectrum (np.array): beat spectrum for the audio file

        Example:

        .. code-block:: python
            :linenos:
            
            # Set up audio signal
            signal = nussl.AudioSignal('path_to_file.wav')

            # Set up a Repet object
            repet = nussl.Repet(signal)

            # I don't have to run repet to get a beat spectrum for signal
            beat_spec = repet.get_beat_spectrum()
            
        """
        # TODO: Make this multi-channel. The np.mean() reduces the n channels to 1.       
        self.beat_spectrum = self.compute_beat_spectrum(
            np.mean(np.square(self.magnitude_spectrogram),
                    axis=constants.STFT_CHAN_INDEX).T
        )
        return self.beat_spectrum

    def _calculate_repeating_period(self):
        # user provided a period, so no calculations to do
        if self.period is not None:
            return self.period

        # get beat spectrum
        self.beat_spectrum = self.get_beat_spectrum()

        # update the min and max so they're in units of time bin indices
        if not self._is_period_converted_to_hops:
            self.min_period = self._update_period(self.min_period)
            self.max_period = self._update_period(self.max_period)
            self._is_period_converted_to_hops = True

        self.repeating_period = self.find_repeating_period_simple(self.beat_spectrum,
                                                                  self.min_period, self.max_period)

        return self.repeating_period

    @staticmethod
    def compute_beat_spectrum(power_spectrogram):
        """ Computes the beat spectrum averages (over freq's) the autocorrelation matrix of a one-sided spectrogram.

        The autocorrelation matrix is computed by taking the autocorrelation of each row of the spectrogram and
        dismissing the symmetric half.

        Args:
            power_spectrogram (:obj:`np.array`): 2D matrix containing the one-sided power spectrogram of an audio signal
            
        Returns:
            (:obj:`np.array`): array containing the beat spectrum based on the power spectrogram
            
        See Also:
            J Foote's original derivation of the Beat Spectrum: 
            Foote, Jonathan, and Shingo Uchihashi. "The beat spectrum: A new approach to rhythm analysis." 
            Multimedia and Expo, 2001. ICME 2001. IEEE International Conference on. IEEE, 2001.
            (`See PDF here <http://rotorbrain.com/foote/papers/icme2001.pdf>`_)
            
        """
        freq_bins, time_bins = power_spectrogram.shape

        # row-wise autocorrelation according to the Wiener-Khinchin theorem
        power_spectrogram = np.vstack([power_spectrogram, np.zeros_like(power_spectrogram)])

        nearest_power_of_two = 2 ** np.ceil(np.log(power_spectrogram.shape[0]) / np.log(2))
        pad_amount = int(nearest_power_of_two - power_spectrogram.shape[0])
        power_spectrogram = np.pad(power_spectrogram, ((0, pad_amount), (0, 0)), 'constant')
        fft_power_spec = scifft.fft(power_spectrogram, axis=0)
        abs_fft = np.abs(fft_power_spec) ** 2
        autocorrelation_rows = np.real(
            scifft.ifft(abs_fft, axis=0)[:freq_bins, :])  # ifft over columns

        # normalization factor
        norm_factor = np.tile(np.arange(freq_bins, 0, -1), (time_bins, 1)).T
        autocorrelation_rows = autocorrelation_rows / norm_factor

        # compute the beat spectrum
        beat_spectrum = np.mean(autocorrelation_rows, axis=1)
        # average over frequencies

        return beat_spectrum

    @staticmethod
    def find_repeating_period_simple(beat_spectrum, min_period, max_period):
        """
        Computes the repeating period of the sound signal using the beat spectrum.
        This algorithm just looks for the max value in the interval 
        ``[min_period, max_period]``, inclusive. It discards the first value, and 
        returns the period in units of stft time bins.

        Args:
            beat_spectrum (:obj:`np.array`): input beat spectrum array
            min_period (int): minimum possible period value
            max_period (int): maximum possible period value
            
        Returns:
             period (int): The period of the sound signal in stft time bins
             
        See Also:
            :func:`find_repeating_period_complex`
            
        """
        min_period, max_period = int(min_period), int(max_period)
        # discard the first element of beat_spectrum (lag 0)
        beat_spectrum = beat_spectrum[1:]
        beat_spectrum = beat_spectrum[min_period - 1: max_period]

        if len(beat_spectrum) == 0:
            raise SeparationException('min_period is larger than the beat spectrum!')

        period = np.argmax(beat_spectrum) + min_period

        return period

    def _compute_repeating_mask(self, magnitude_spectrogram_channel):
        """
        Computes the soft mask for the repeating part using the magnitude 
        spectrogram and the repeating period

        Args:
            magnitude_spectrogram_channel (:obj:`np.array`): 2D matrix containing the 
              magnitude spectrogram of a signal
            
        Returns:
            (:obj:`np.array`): 2D matrix (Lf by Lt) containing the soft mask for the 
              repeating part, elements of M take on values in ``[0, 1]``

        """
        period = self.repeating_period
        freq_bins, time_bins = magnitude_spectrogram_channel.shape
        n_repetitions = int(np.ceil(float(time_bins) / period))
        one_period = freq_bins * period

        # Pad to make an integer number of repetitions. Pad with 'nan's to not affect the median.
        remainder = (period * n_repetitions) % time_bins
        mask_reshaped = np.hstack(
            [magnitude_spectrogram_channel,
             float('nan') * np.zeros((freq_bins, remainder))]
        )

        # reshape to take the median of each period
        mask_reshaped = np.reshape(mask_reshaped.T, (n_repetitions, one_period))

        # take median of repeating periods before and after the padding
        median_mask = np.nanmedian(mask_reshaped, axis=0)

        # reshape to it's original shape
        median_mask = np.reshape(
            np.tile(median_mask, (n_repetitions, 1)),
            (n_repetitions * period, freq_bins)).T
        median_mask = median_mask[:, :time_bins]

        # take minimum of computed mask and original input and scale
        min_median_mask = np.minimum(median_mask, magnitude_spectrogram_channel)
        mask = (min_median_mask + constants.EPSILON) / (
                    magnitude_spectrogram_channel + constants.EPSILON)

        return mask

    def _update_period(self, period):
        period = float(period)
        secs_per_bin = self.stft_params.hop_length / self.audio_signal.sample_rate
        bins_in_period = period / secs_per_bin
        period_in_frames = int(np.ceil(bins_in_period))
        return period_in_frames



import warnings

import torch
import torch.nn as nn
import librosa
import numpy as np
from torch.utils.checkpoint import checkpoint

class AmplitudeToDB(nn.Module):
    """
    Takes a magnitude spectrogram and converts it to a log
    amplitude spectrogram in decibels.
    
    Args:
        data (torch.Tensor): Magnitude spectrogram to convert to
        log spectrogram.
        ref (float): reference value. Defaults to 1.0.
        amin (float): lowest possible value for numerical stability.
        Defaults to 1e-8.
    
    Returns:
        torch.Tensor: log-magnitude energy in each bin in representation.
    """

    def forward(self, data, ref=1.0, amin=1e-4):
        data = data ** 2
        amin = amin ** 2
        ref = np.log10(np.maximum(amin, ref ** 2))
        data = 10.0 * (torch.log10(torch.clamp(data, min=amin)) - ref)
        return data

class ShiftAndScale(nn.Module):
    """
    Takes in some data, shifts it by a learned bias and multiplies it
    by a learned scalar. 
    
    Args:
        torch.Tensor: shifted and scaled data
    """
    def __init__(self, learn_scale=True, learn_shift=True):
        super().__init__()
        self.scale = nn.Parameter(
            torch.ones(1), requires_grad=learn_scale)
        self.shift = nn.Parameter(
            torch.zeros(1), requires_grad=learn_shift)

    def forward(self, data):
        return self.scale * data + self.shift

class BatchNorm(nn.Module):
    """
    Applies a batch norm layer. Defaults to using only 1 feature, commonly
    used at the very beginning of the network to normalize the input spectrogram.

    Data comes and goes as (nb, nt, nf, nc, ...). Inside this module, the data undergoes
    the following procedure:

    1. It is first reshaped to (nb, nf, nt, nc, ...)
    2. Data is reshaped to (nb, nf, -1).
    3. ``BatchNorm1d`` is applied with ``num_features`` to data.
    4. Data is reshaped back to (nb, nt, nf, nc) and returned.
    
    Args:
        num_features (int): num_features argument to BatchNorm1d, defaults to 1.
        feature_dim (int): which dimension the features that are being normalized are on.
          Defaults to 2.
        **kwargs (dict): additional keyword arguments that can be passed to BatchNorm2d.
    
    Returns:
        torch.Tensor: modified input data tensor with batch norm.
    """

    def __init__(self, num_features=1, feature_dim=2, **kwargs):
        super(BatchNorm, self).__init__()
        self.num_features = num_features
        self.feature_dim = feature_dim
        self.add_module('batch_norm', nn.BatchNorm1d(self.num_features, **kwargs))

    def forward(self, data):
        data = data.transpose(self.feature_dim, 1)
        shape = data.shape
        new_shape = (shape[0], self.num_features, -1)

        data = data.reshape(new_shape)
        data = self.batch_norm(data)
        data = data.reshape(shape)
        data = data.transpose(self.feature_dim, 1)
        return data


class InstanceNorm(nn.Module):
    """
    Applies an instance norm layer. Defaults to using only 1 feature, commonly
    used at the very beginning of the network to normalize the input spectrogram.

    Data comes and goes as (nb, nt, nf, nc). Inside this module, the data undergoes
    the following procedure:

    1. It is first reshaped to (nb, nf, nt, nc)
    2. Data is reshaped to (nb, nf, nt * nc).
    3. ``InstanceNorm1d`` is applied with ``num_features`` to data.
    4. Data is reshaped back to (nb, nt, nf, nc) and returned.
    
    Args:
        num_features (int): num_features argument to InstanceNorm1d, defaults to 1.
        feature_dim (int): which dimension the features that are being normalized are on.
          Defaults to 2.
        **kwargs (dict): additional keyword arguments that are passed to InstanceNorm2d.
    
    Returns:
        torch.Tensor: modified input data tensor with instance norm applied.
    """

    def __init__(self, num_features=1, feature_dim=2, **kwargs):
        super(InstanceNorm, self).__init__()
        self.num_features = num_features
        self.feature_dim = feature_dim
        self.add_module('instance_norm', nn.InstanceNorm1d(self.num_features, **kwargs))

    def forward(self, data):
        data = data.transpose(self.feature_dim, 1)
        shape = data.shape
        new_shape = (shape[0], self.num_features, -1)

        data = data.reshape(new_shape)
        data = self.instance_norm(data)
        data = data.reshape(shape)
        data = data.transpose(self.feature_dim, 1)
        return data

class GroupNorm(nn.Module):
    """
    Applies a group norm layer. Defaults to using 1 group and all feature, commonly
    used at the very beginning of the network to normalize each feature of
    the input spectrogram.

    Data comes and goes as (nb, nt, nf, nc). Inside this module, the data undergoes
    the following procedure:

    1. It is first reshaped to (nb, nf, nt, nc)
    2. Data is reshaped to (nb, nf, nt * nc).
    3. ``GroupNorm`` is applied with ``num_features`` to data.
    4. Data is reshaped back to (nb, nt, nf, nc) and returned.
    
    Args:
        num_features (int): num_features argument to GroupNorm.
        num_groups (int): Number of groups, defaults to 1.
        feature_dim (int): which dimension the features that are being normalized are on.
          Defaults to 2.
        **kwargs (dict): additional keyword arguments that are passed to InstanceNorm2d.
    
    Returns:
        torch.Tensor: modified input data tensor with instance norm applied.
    """

    def __init__(self, num_features, num_groups=1, feature_dim=2, **kwargs):
        super().__init__()
        self.num_features = num_features
        self.num_groups = num_groups
        self.feature_dim = feature_dim
        self.add_module('group_norm', nn.GroupNorm(
            self.num_groups, self.num_features, **kwargs))

    def forward(self, data):
        data = data.transpose(self.feature_dim, 1)
        shape = data.shape
        new_shape = (shape[0], self.num_features, -1, 1)

        data = data.reshape(new_shape)
        data = self.group_norm(data)
        data = data.reshape(shape)
        data = data.transpose(self.feature_dim, 1)
        return data

class Alias(nn.Module):
    """
    Super simple module that just passes the data through without altering it, so
    that the output of a model can be renamed in a SeparationModel.
    """
    def forward(self, data):
        return data

class LayerNorm(nn.Module):
    """
    Applies an layer norm layer. Defaults to using only 1 feature, commonly
    used at the very beginning of the network to normalize the input representation.
    This wraps nn.LayerNorm with some quality of life improvements for dealing
    specifically with audio representations.

    Data comes and goes as (nb, nt, nf, nc). Inside this module, the data undergoes
    the following procedure:

    1. It is first transposed according to feature_dims, moving relevant features
       to the end. For example, (nb, nt, nf, nc) -> (nb, nt, nc, nf), with 
       normalized_shape = (nf,).
    2. ``LayerNorm`` is applied with ``normalized_shape`` to data.
    3. Data is transposed and reshaped back to (nb, nt, nf, nc) and returned.
    
    Args:
        normalized_shape (int, list, or tuple): normalized_shape argument to LayerNorm.
        feature_dims (list): The dimensions that the features are on, should be
          the same length as normalized_shape. These features will be moved to the end
          of the tensor in order, then layer norm is applied, then the features are 
          moved back to their original locations.
        kwargs: additional keyword arguments that are passed to nn.LayerNorm.
    
    Returns:
        torch.Tensor: modified input data tensor with layer norm applied.
    """

    def __init__(self, num_features, feature_dims, **kwargs):
        super().__init__()
        
        self.gamma = nn.Parameter(torch.ones(num_features))
        self.beta = nn.Parameter(torch.zeros(num_features))
        self.feature_dims = feature_dims

    def forward(self, data):
        # make a bunch of dummy dimensions to move
        # things to the end
        mean_dims = []
        for d in self.feature_dims:
            data = data.unsqueeze(-1).transpose(d, -1)
            mean_dims.append(len(data.shape) - 1)
        mean = data.mean(dim=mean_dims, keepdims=True)
        var = data.var(dim=mean_dims, keepdims=True)
        data = (data - mean) / ((var + 1e-8) ** 0.5)
        data = data * self.gamma + self.beta
        # move everything back and squeeze off extra 
        # dimensions
        for d in self.feature_dims[::-1]:
            data = data.transpose(-1, d).squeeze(-1)
        return data

class MelProjection(nn.Module):
    """
    MelProjection takes as input a time-frequency representation (e.g. a spectrogram, or a mask) and outputs a mel
    project that can be learned or fixed. The initialization uses librosa to get a mel filterbank. Direction
    controls whether it is a forward transform or the inverse transform (e.g. back to spectrogram).

    Args:
        sample_rate: (int) Sample rate of audio for computing the mel filters.
        num_frequencies: (int) Number of frequency bins in input spectrogram.
        num_mels: (int) Number of mel bins in output mel spectrogram. if num_mels < 0, this does nothing
            other than clamping the output if clamp is True.
        direction: (str) Which direction to go in (either 'forward' - to mel, or 'backward' - to frequencies).
            Defaults to 'forward'.
        clamp: (bool) Whether to clamp the output values of the transform between 0.0 and 1.0. Used for transforming
            a mask in and out of the mel-domain. Defaults to False.
        trainable: (bool) Whether the mel transform can be adjusted by the optimizer. Defaults to False.
    """
    def __init__(self, sample_rate, num_frequencies, num_mels, direction='forward',
                 clamp=False, trainable=False):
        super(MelProjection, self).__init__()
        self.num_mels = num_mels
        if direction not in ['backward', 'forward']:
            raise ValueError("direction must be one of ['backward', 'forward']!")
        self.direction = direction
        self.clamp = clamp

        if self.num_mels > 0:
            shape = (
                (num_frequencies, num_mels)
                if self.direction == 'forward' else
                (num_mels, num_frequencies)
            )
            self.add_module('transform', nn.Linear(*shape))

            mel_filters = librosa.filters.mel(sample_rate, 2 * (num_frequencies - 1), num_mels)
            mel_filters = (mel_filters.T / (mel_filters.sum(axis=1) + 1e-8)).T
            filter_bank = (
                mel_filters
                if self.direction == 'forward'
                else np.linalg.pinv(mel_filters)
            )

            for name, param in self.transform.named_parameters():
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                if 'weight' in name:
                    param.data = torch.from_numpy(filter_bank).float()
                param.requires_grad_(trainable)

    def forward(self, data):
        """
        Args:
            data: Representation - shape: 
              (batch_size, sequence_length, num_frequencies or num_mels, num_sources)

        Returns:
            Mel-spectrogram or time-frequency representation of shape:
              (batch_size, sequence_length, num_mels or num_frequencies, num_sources).
        """

        if self.num_mels > 0:
            data = data.transpose(2, -1)
            data = self.transform(data)
            data = data.transpose(2, -1)

        if self.clamp:
            data = data.clamp(0.0, 1.0)
        return data.contiguous()


class Embedding(nn.Module):
    """
    Maps output from an audio representation module (e.g. RecurrentStack, 
    DilatedConvolutionalStack) to an embedding space. The output shape is 
    (batch_size, sequence_length, num_features, embedding_size). The embeddings can
    be passed through an activation function. If activation is 'softmax' or 
    'sigmoid', and embedding_size is equal to the number of sources, this module 
    can be used to implement a mask inference network (or a mask inference
    head in a Chimera network setup).

    Args:
        num_features (int): Number of features being mapped for each frame. 
          Either num_frequencies, or if used with MelProjection, num_mels if using 
          RecurrentStack. Should be 1 if using DilatedConvolutionalStack. 

        hidden_size (int): Size of output from RecurrentStack (hidden_size) or 
          DilatedConvolutionalStack (num_filters). If RecurrentStack is bidirectional, 
          this should be set to 2 * hidden_size.
        
        embedding_size (int): Dimensionality of embedding.
        
        activation (list of str): Activation functions to be applied. Options 
          are 'sigmoid', 'tanh', 'softmax', 'relu', 'gate'. Unit normalization can be applied by 
          adding 'unit_norm' to the list (e.g. ['sigmoid', unit_norm']). Alternatively,
          L1 normalization can be applied by adding 'l1_norm' to the list.

        dim_to_embed (int): Which dimension of the input to apply the embedding to.
          Defaults to -1 (the last dimension).

        bias (bool): Whether or not to place a bias on the linear layer. Defaults to
          True.

        reshape (bool): Whether to reshape the output of the linear layer to look
          like a time-frequency representation (nb, nt, nf, nc, ...). Defaults to
          True.
    """
    def __init__(self, num_features, hidden_size, embedding_size, activation,
                 num_audio_channels=1, dim_to_embed=-1, bias=True, reshape=True):
        super(Embedding, self).__init__()
        self.add_module(
            'linear',
            nn.Linear(
                hidden_size, 
                num_features * num_audio_channels * embedding_size,
                bias=bias
            )
        )
        self.num_features = num_features
        self.num_audio_channels = num_audio_channels
        self.activation = activation
        self.embedding_size = embedding_size
        self.reshape = reshape

        if 'gated_tanh' in self.activation:
            self.embed_linear = nn.Sequential(
                nn.Linear(hidden_size, hidden_size), 
                nn.Tanh()
            )
            self.embed_gate = nn.Sequential(
                nn.Linear(hidden_size, hidden_size),
                nn.Sigmoid()
            )

        if isinstance(dim_to_embed, int):
            dim_to_embed = [dim_to_embed]

        self.dim_to_embed = dim_to_embed

        for name, param in self.linear.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0.0)
            elif 'weight' in name:
                nn.init.xavier_normal_(param)

    def forward(self, data):
        """
        Args:
            data: output from RecurrentStack or ConvolutionalStack. Shape is:
              (num_batch, ..., hidden_size or num_filters)

        Returns:
            An embedding (with an optional activation) for each point in the 
              representation of shape (num_batch, ..., embedding_size).
        """
        shape = list(data.shape)
        _dims = []
        for _dim in self.dim_to_embed:
            # move each dimension to embed to end of tensor
            if _dim == -1:
                _dim = len(shape) - 1
            data = data.transpose(_dim, -1)
            _dims.append(_dim)

        # the new shape doesn't have the embedded dimensions
        shape = [
            v for i, v in enumerate(shape) 
            if i not in _dims
        ]
        
        shape = tuple(shape)
        data = data.reshape(shape + (-1,))

        if 'gated_tanh' in self.activation:
            data = self.embed_linear(data) * self.embed_gate(data)

        data = self.linear(data)

        if self.reshape:
            shape = shape + (
                self.num_features, self.num_audio_channels, self.embedding_size,)
            data = data.reshape(shape)

        if 'sigmoid' in self.activation:
            data = torch.sigmoid(data)
        elif 'tanh' in self.activation:
            data = torch.tanh(data)
        elif 'relu' in self.activation:
            data = torch.relu(data)
        elif 'softmax' in self.activation:
            data = torch.softmax(data, dim=-1)

        if 'unit_norm' in self.activation:
            data = nn.functional.normalize(data, dim=-1, p=2)
        elif 'l1_norm' in self.activation:
            data = nn.functional.normalize(data, dim=-1, p=1)

        return data


class Mask(nn.Module):
    """
    Takes a mask and applies it to a representation. Mask and representation must match
    in their first 3 dimensions (nb, nt, nf). The last
    dimension of the representation is unsqueezed to match the mask shape. So if there
    are ``ns`` sources to separate, the mask shape will be (nb, nt, nf, ns), the 
    representation shape will be (nb, nt, nf).

    Representation gets unsqueezed to (nb, nt, nf, 1). Multiplying with the mask
    broadcasts, resulting in (nb, nt, nf, ns) output corresponding to each separated
    source from the representation.

    Detaches the representation before multiplying the mask element-wise.
    """

    def __init__(self):
        super(Mask, self).__init__()

    def forward(self, mask, representation):
        # add a source dimension
        representation = representation.unsqueeze(-1).expand_as(mask)
        return mask * representation.detach()

class Split(nn.Module):
    """
    Splits a piece of data on a specified axis into a tuple of
    specified splits. Defaults to splitting on the last axis.

    Example:
    
    .. code-block:: python

        data = torch.randn(100, 10)
        split = Split((3, 7), dim=-1)
        split1, split2 = split(data)

        split1 # has shape (100, 3)
        split2 # has shape (100, 7)

    """
    def __init__(self, split_sizes, dim=-1):
        self.dim = dim
        self.split_sizes = split_sizes
        super().__init__()

    def forward(self, data):
        data = torch.split_with_sizes(
            data, self.split_sizes, dim=self.dim)
        return data

class Expand(nn.Module):
    """
    Expand a seconod tensor to match the dimensions of the first
    tensor, if possible. Just wraps `.expand_as`. Expands on
    the last axis.
    """
    def __init__(self):
        super().__init__()
    
    def forward(self, tensor_a, tensor_b):
        if tensor_a.ndim < tensor_b.ndim:
            raise ValueError("tensor_a should have more dimensions than tensor b!")
        for _ in range(tensor_a.ndim - tensor_b.ndim):
            tensor_b = tensor_b.unsqueeze(-1)
        return tensor_b.expand_as(tensor_a)


class Concatenate(nn.Module):
    """
    Concatenates two or more pieces of data together along a 
    specified dimension. Takes in a list of tensors and a 
    concatenation dimension.
    """
    def __init__(self, dim=-1):
        self.dim = dim
        super(Concatenate, self).__init__()
    
    def forward(self, *data):
        return torch.cat(data, dim=self.dim)

class RecurrentStack(nn.Module):
    """
    Creates a stack of RNNs used to process an audio sequence represented as 
    (sequence_length, num_features). With bidirectional = True, hidden_size = 600, 
    num_layers = 4, rnn_type='lstm', and dropout = .3, this becomes the 
    audio processor used in deep clustering networks, deep attractor networks, etc. 
    Note that batch_first is set to True here.

    Args:
        num_features: (int) Number of features being mapped for each frame. 
            Either num_frequencies, or if used with MelProjection, num_mels.
        hidden_size: (int) Hidden size of recurrent stack for each layer.
        num_layers: (int) Number of layers in stack.
        bidirectional: (int) True makes this a BiLSTM or a BiGRU. Note that this 
            doubles the hidden size.
        dropout: (float) Dropout between layers.
        rnn_type: (str) LSTM ('lstm') or GRU ('gru').
    """
    def __init__(self, num_features, hidden_size, num_layers, bidirectional, dropout,
                 rnn_type='lstm', batch_first=True, init_forget=True):
        super(RecurrentStack, self).__init__()
        if rnn_type not in ['lstm', 'gru']:
            raise ValueError("rnn_type must be one of ['lstm', 'gru']!")

        RNNClass = nn.LSTM if rnn_type == 'lstm' else nn.GRU
        self.add_module(
            'rnn', RNNClass(
                num_features, hidden_size, num_layers, batch_first=batch_first,
                bidirectional=bidirectional, dropout=dropout))

        if init_forget:
            for name, param in self.rnn.named_parameters():
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_normal_(param)

            for names in self.rnn._all_weights:
                for name in filter(lambda nm: "bias" in nm, names):
                    bias = getattr(self.rnn, name)
                    n = bias.size(0)
                    start, end = n // 4, n // 2
                    bias.data[start:end].fill_(1.)

    def forward(self, data):
        """
        Args:
            data: Audio representation to be processed. Should be of shape:
            (num_batch, sequence_length, ...).

        Returns:
            Outputs the features after processing of the RNN. Shape is:
            (num_batch, sequence_length, hidden_size or hidden_size*2 if 
            bidirectional=True)
        """
        shape = data.shape
        data = data.reshape(shape[0], shape[1], -1)
        self.rnn.flatten_parameters()
        data = self.rnn(data)[0]
        return data


class ConvolutionalStack2D(nn.Module):
    """
    Implements a stack of dilated convolutional layers for source separation from 
    the following papers:

        Mobin, Shariq, Brian Cheung, and Bruno Olshausen. "Convolutional vs. recurrent 
        neural networks for audio source separation." 
        arXiv preprint arXiv:1803.08629 (2018). https://arxiv.org/pdf/1803.08629.pdf

        Yu, Fisher, Vladlen Koltun, and Thomas Funkhouser. "Dilated residual networks." 
        Proceedings of the IEEE conference on computer vision and pattern recognition. 
        2017. https://arxiv.org/abs/1705.09914
    
    Args:
        in_channels (int): Number of channels in input
        channels (list of int): Number of channels for each layer
        dilations (list of ints or int tuples): Dilation rate for each layer. If 
            int, it is same in both height and width. If tuple, tuple is defined as
            (height, width). 
        filter_shapes (list of ints or int tuples): Filter shape for each layer. If 
            int, it is same in both height and width. If tuple, tuple is defined as
            (height, width).
        residuals (list of bool): Whether or not to keep a residual connection at
            each layer.
        batch_norm (bool): Whether to use BatchNorm or not at each layer (default: True)
        use_checkpointing (bool): Whether to use torch's checkpointing functionality 
            to reduce memory usage.
    
    Raises:
        ValueError -- All the input lists must be the same length.
    """
    def __init__(self, in_channels, channels, dilations, filter_shapes, residuals,
                 batch_norm=True, use_checkpointing=False):
        for x in [dilations, filter_shapes, residuals]:
            if len(x) != len(channels):
                raise ValueError(
                    f"All lists (channels, dilations, filters, residuals) should have"
                    f"the same length!"
                )
        super().__init__()

        if any([d != 1 for d in dilations]):
            warnings.warn(
                "You specified a dilation != 1. Input size and output size are "
                "not guaranteed to be the same! This is due to the lacking of "
                "padding = 'same' in PyTorch."
            )

        self.dilations = dilations
        self.filter_shapes = filter_shapes
        self.residuals = residuals
        self.batch_norm = batch_norm
        self.padding = None
        self.channels = channels
        self.in_channels = in_channels
        self.use_checkpointing = use_checkpointing
        self.num_layers = len(channels)
        self.layers = nn.ModuleList()

        for i in range(len(channels)):
            self.layers.append(self._make_layer(i))

    def _make_layer(self, i):
        convolution = nn.Conv2d(
            in_channels=self.channels[i - 1] if i > 0 else self.in_channels,
            out_channels=self.channels[i],
            kernel_size=self.filter_shapes[i],
            dilation=self.dilations[i],
            padding=self.filter_shapes[i] // 2,
        )

        if i == len(self.channels) - 1:
            layer = convolution
            self.add_module(f'layer{i}', layer)
            return layer

        layer = nn.Sequential()
        layer.add_module('conv', convolution)
        if self.batch_norm:
            batch_norm = nn.BatchNorm2d(self.channels[i])
            layer.add_module('batch_norm', batch_norm)
        layer.add_module('relu', nn.ReLU())
        return layer

    def layer_function(self, data, layer, previous_layer, i):
        if self.use_checkpointing:
            data = checkpoint(layer, data)
        else:
            data = layer(data)
        if self.residuals[i]:
            if previous_layer is not None:
                data += previous_layer
            previous_layer = data
        return data, previous_layer

    def forward(self, data):
        """ 
        Data comes in as: [num_batch, sequence_length, num_frequencies, num_audio_channels]
        We reshape it in the forward pass so that everything works to:
          [num_batch, num_audio_channels, sequence_length, num_frequencies]
        After this input is processed, the shape is then:
          [num_batch, num_output_channels, sequence_length, num_frequencies]
        We transpose again to make the shape:
          [num_batch, sequence_length, num_frequencies, num_output_channels]
        So it can be passed to an Embedding module.
        """
        data = data.permute(0, 3, 1, 2)
        previous_layer = None
        for i in range(self.num_layers):
            data, previous_layer = self.layer_function(
                data, self.layers[i], previous_layer, i
            )
        data = data.permute(0, 2, 3, 1)
        return data

def _find_and_instantiate_module(module_name, module_args):
    # Check this file first
    module = None
    if module_name in globals():
        module = globals()[module_name](**module_args)
    # Then check torch.nn
    elif module_name in dir(nn):
        module_func = getattr(nn, module_name)
        module = module_func(**module_args)
    return module

class DualPathBlock(nn.Module):
    def __init__(self, n_features, hidden_size, 
                 intra_processor, inter_processor, 
                 **kwargs):
        super().__init__()
        # Saving needed stuff
        self.n_features = n_features
        self.hidden_size = hidden_size
        
        # Projections down and up
        self.intra_fc = nn.Linear(
            hidden_size, n_features)
        self.inter_fc = nn.Linear(
            hidden_size, n_features)
        
        # Inter/intra chunk processors
        if isinstance(inter_processor, dict):
            inter_processor = _find_and_instantiate_module(
                inter_processor['class'], inter_processor['args'])
        if isinstance(intra_processor, dict):
            intra_processor = _find_and_instantiate_module(
                intra_processor['class'], intra_processor['args'])
        
        if any([intra_processor is None, inter_processor is None]): 
            raise ValueError(
                f"Tried to instantiate inter/intra processor, but failed!")

        self.intra_processor = intra_processor
        self.inter_processor = inter_processor

        # Normalization + activation layers
        self.intra_norm = nn.GroupNorm(1, n_features, eps=1e-8)
        self.inter_norm = nn.GroupNorm(1, n_features, eps=1e-8)

    def apply_norm(self, data, norm_layer, shape):
        in_shape = data.shape
        data = data.reshape(*shape)
        # data is (batch_size, n_chunks, chunk_size, n_features)
        data = data.permute(0, 3, 1, 2)
        # now (nb, nf, nch, ncs)
        data = norm_layer(data)
        data = data.permute(0, 2, 3, 1)
        data = data.reshape(*in_shape)
        return data
        
            
    def forward(self, data):
        # data comes in as (batch_size, n_chunks, chunk_size, n_features)
        batch_size, n_chunks, chunk_size, n_features = data.shape
        data = data.reshape(batch_size * n_chunks, chunk_size, data.shape[-1])

        output = data # Skip connection -------->--------->
        # Apply intra-chunk processor                   # | 
        data = self.intra_processor(data)               # | 
        data = self.intra_fc(data)                      # |
        data = self.apply_norm(                         # |
            data, self.intra_norm, (                    # |
                batch_size, n_chunks, chunk_size, -1    # |
            )                                           # |
        )                                               # |
        data = output + data  # <---------------<---------<

        # Reshape so chunks are on other side
        data = data.reshape(batch_size, n_chunks, chunk_size, -1)
        data = data.transpose(2, 1)
        data = data.reshape(batch_size * chunk_size, n_chunks, -1)

        output = data # Skip connection -------->--------->
        # Apply inter-chunk processor                   # | 
        data = self.inter_processor(data)               # | 
        data = self.inter_fc(data)                      # |
        data = self.apply_norm(                         # |
            data, self.inter_norm, (                    # |
                batch_size, chunk_size, n_chunks, -1    # |
            )                                           # |
        )                                               # |
        data = output + data  # <---------------<---------<
        # Reshape back to original
        data = data.reshape(batch_size, chunk_size, n_chunks, -1)
        data = data.transpose(2, 1)
        return data
    
class DualPath(nn.Module):
    """
    Implements a dual path processor, which takes a sequence and splits it 
    into overlapping chunks. Each chunk is then processed by a specified
    class which will process each chunk individually, and also process
    the chunks in sequence. The output of the processor then resynthesizes
    something with the input shape by doing an overlap/add procedure.
    
    Args:
        num_layers (int): Number of layers.
        chunk_size (int): Chunk size of intra-processing.
        hop_size (int): Hop length between chunks.
        layer_class (object): Class to use to process the data. Defaults to
          DualPathBlock. Keyword arguments can be specified.
        skip_connection (bool, optional): Whether to put skip connections between
          each layer. Defaults to False.
        kwargs: Keyword arguments to the layer class.
    """
    def __init__(self, num_layers, chunk_size, hop_size, in_features,
                 bottleneck_size, skip_connection=False, **kwargs):
        super().__init__()
        
        self.chunk_size = chunk_size
        self.hop_size = hop_size
        
        self.layers = nn.ModuleList()
        for i in range(num_layers):
            _block = DualPathBlock(n_features=bottleneck_size, **kwargs)
            self.layers.append(_block)
        self.skip_connection = skip_connection
        self.prelu = nn.PReLU()
        self.bottleneck = nn.Linear(in_features, bottleneck_size)
        self.bottleneck_norm = nn.GroupNorm(1, in_features)
        self.inv_bottleneck = nn.Linear(
            bottleneck_size, in_features)
    
    def forward(self, data):
        fold = nn.Fold(
            (1, data.shape[1]), 
            (1, self.chunk_size), 
            stride=(1, self.hop_size)
        )
        unfold = nn.Unfold(
            (1, self.chunk_size), stride=(1, self.hop_size)
        )

        batch_size, sequence_length, n_features, n_channels = (
            data.shape
        )
        
        # extract chunks
        data = data.transpose(1, 2)
        data = self.bottleneck_norm(data)
        data = data.transpose(1, -1)
        data = self.bottleneck(data)
        data = data.permute(0, 2, 3, 1)

        data = data.reshape(
            batch_size, sequence_length, 1, -1)

        data = data.transpose(3, 1)
        data = unfold(data)
        # unfold makes the data (batch_size, bottleneck_size * chunk_size, n_chunks)
        n_chunks = data.shape[-1]
        data = data.reshape(batch_size, -1, self.chunk_size, n_chunks)
        data = data.transpose(3, 1)
        # data is now (batch_size, n_chunks, chunk_size, bottleneck_size)
        # process
        output = data # Skip connection --->
        for layer in self.layers:        # | 
            data = layer(output)         # |
            if self.skip_connection:     # |
                output += data # <----<----<
            else:
                output = data
        data = output
        data = self.prelu(data)
        # data is still (batch_size, n_chunks, chunk_size, bottleneck_size)
        data = self.inv_bottleneck(data)
        # data is now (batch_size, n_chunks, chunk_size, in_features)
        #         
        # resynthesize with overlap/add
        data = data.transpose(1, 3)
        data = data.reshape(-1, n_features * self.chunk_size, n_chunks)
        data = fold(data)
        data = data.transpose(3, 1)

        data = data.reshape(
            batch_size, sequence_length, n_features, n_channels)        
        
        return data


from itertools import permutations, combinations

import torch
import torch.nn as nn

class L1Loss(nn.L1Loss):
    DEFAULT_KEYS = {'estimates': 'input', 'source_magnitudes': 'target'}

class MSELoss(nn.MSELoss):
    DEFAULT_KEYS = {'estimates': 'input', 'source_magnitudes': 'target'}

class KLDivLoss(nn.KLDivLoss):
    DEFAULT_KEYS = {'estimates': 'input', 'source_magnitudes': 'target'}

class SISDRLoss(nn.Module):
    """
    Computes the Scale-Invariant Source-to-Distortion Ratio between a batch
    of estimated and reference audio signals. Used in end-to-end networks.
    This is essentially a batch PyTorch version of the function 
    ``nussl.evaluation.bss_eval.scale_bss_eval`` and can be used to compute
    SI-SDR or SNR.
    
    Args:
        scaling (bool, optional): Whether to use scale-invariant (True) or
          signal-to-noise ratio (False). Defaults to True.
        return_scaling (bool, optional): Whether to only return the scaling
          factor that the estimate gets scaled by relative to the reference. 
          This is just for monitoring this value during training, don't actually
          train with it! Defaults to False.
        reduction (str, optional): How to reduce across the batch (either 'mean', 
          'sum', or none). Defaults to 'mean'.
        zero_mean (bool, optional): Zero mean the references and estimates before
          computing the loss. Defaults to True.
    """
    DEFAULT_KEYS = {'audio': 'estimates', 'source_audio': 'references'}

    def __init__(self, scaling=True, return_scaling=False, reduction='mean', 
                 zero_mean=True):
        self.scaling = scaling
        self.reduction = reduction
        self.zero_mean = zero_mean
        self.return_scaling = return_scaling
        super().__init__()

    def forward(self, estimates, references):
        eps = 1e-8
        # num_batch, num_samples, num_sources
        _shape = references.shape
        references = references.reshape(-1, _shape[-2], _shape[-1])
        estimates = estimates.reshape(-1, _shape[-2], _shape[-1])
        # samples now on axis 1
        if self.zero_mean:
            mean_reference = references.mean(dim=1, keepdim=True)
            mean_estimate = estimates.mean(dim=1, keepdim=True)
        else:
            mean_reference = 0
            mean_estimate = 0
        
        _references = references - mean_reference
        _estimates = estimates - mean_estimate
        
        references_projection = (_references ** 2).sum(dim=-2) + eps
        references_on_estimates = (_estimates * _references).sum(dim=-2) + eps

        scale = (
            (references_on_estimates / references_projection).unsqueeze(1)
            if self.scaling else 1)

        e_true = scale * _references
        e_res = _estimates - e_true

        signal = (e_true ** 2).sum(dim=1)
        noise = (e_res ** 2).sum(dim=1)
        sdr = 10 * torch.log10(signal / noise + eps)        

        if self.reduction == 'mean':
            sdr = sdr.mean()
        elif self.reduction == 'sum':
            sdr = sdr.sum()
        if self.return_scaling:
            return scale
        # go negative so it's a loss
        return -sdr


class DeepClusteringLoss(nn.Module):
    """
    Computes the deep clustering loss with weights. Equation (7) in [1].

    References:

    [1] Wang, Z. Q., Le Roux, J., & Hershey, J. R. (2018, April).
        Alternative Objective Functions for Deep Clustering.
        In Proc. IEEE International Conference on Acoustics,  Speech
        and Signal Processing (ICASSP).
    """
    DEFAULT_KEYS = {
        'embedding': 'embedding', 
        'ideal_binary_mask': 'assignments', 
        'weights': 'weights'
    }

    def __init__(self):
        super(DeepClusteringLoss, self).__init__()

    def forward(self, embedding, assignments, weights):
        batch_size = embedding.shape[0]
        embedding_size = embedding.shape[-1]
        num_sources = assignments.shape[-1]

        weights = weights.view(batch_size, -1, 1)

        # make everything unit norm
        embedding = embedding.reshape(batch_size, -1, embedding_size)
        embedding = nn.functional.normalize(embedding, dim=-1, p=2)

        assignments = assignments.view(batch_size, -1, num_sources)
        assignments = nn.functional.normalize(assignments, dim=-1, p=2)

        norm = (((weights.reshape(batch_size, -1)) ** 2).sum(dim=1) ** 2) + 1e-8

        assignments = weights * assignments
        embedding = weights * embedding

        vTv = ((embedding.transpose(2, 1) @ embedding) ** 2).reshape(
            batch_size, -1).sum(dim=-1)
        vTy = ((embedding.transpose(2, 1) @ assignments) ** 2).reshape(
            batch_size, -1).sum(dim=-1)
        yTy = ((assignments.transpose(2, 1) @ assignments) ** 2).reshape(
            batch_size, -1).sum(dim=-1)
        loss = (vTv - 2 * vTy + yTy) / norm
        return loss.mean()

class WhitenedKMeansLoss(nn.Module):
    """
    Computes the whitened K-Means loss with weights. Equation (6) in [1].

    References:

    [1] Wang, Z. Q., Le Roux, J., & Hershey, J. R. (2018, April).
        Alternative Objective Functions for Deep Clustering.
        In Proc. IEEE International Conference on Acoustics,  Speech
        and Signal Processing (ICASSP).
    """
    DEFAULT_KEYS = {
        'embedding': 'embedding', 
        'ideal_binary_mask': 'assignments', 
        'weights': 'weights'
    }

    def __init__(self):
        super(WhitenedKMeansLoss, self).__init__()

    def forward(self, embedding, assignments, weights):
        batch_size = embedding.shape[0]
        embedding_size = embedding.shape[-1]
        num_sources = assignments.shape[-1]

        weights = weights.view(batch_size, -1, 1)

        # make everything unit norm
        embedding = embedding.reshape(batch_size, -1, embedding_size)
        embedding = nn.functional.normalize(embedding, dim=-1, p=2)

        assignments = assignments.view(batch_size, -1, num_sources)
        assignments = nn.functional.normalize(assignments, dim=-1, p=2)

        assignments = weights * assignments
        embedding = weights * embedding

        embedding_dim_identity = torch.eye(
            embedding_size, device=embedding.device).float()
        source_dim_identity = torch.eye(
            num_sources, device=embedding.device).float()

        vTv = (embedding.transpose(2, 1) @ embedding)
        vTy = (embedding.transpose(2, 1) @ assignments)
        yTy = (assignments.transpose(2, 1) @ assignments)

        ivTv = torch.inverse(vTv + embedding_dim_identity)
        iyTy = torch.inverse(yTy + source_dim_identity)

        ivTv_vTy = ivTv @ vTy
        vTy_iyTy = vTy @ iyTy

        # tr(AB) = sum(A^T o B) 
        # where o denotes element-wise product
        # this is the trace trick
        # http://andreweckford.blogspot.com/2009/09/trace-tricks.html
        trace = (ivTv_vTy * vTy_iyTy).sum()
        D = (embedding_size + num_sources) * batch_size
        loss = D - 2 * trace
        return loss / batch_size

class PermutationInvariantLoss(nn.Module):
    """
    Computes the Permutation Invariant Loss (PIT) [1] by permuting the estimated 
    sources and the reference sources. Takes the best permutation and only backprops
    the loss from that.

    For when you're trying to match the estimates to the sources but you don't 
    know the order in which your model outputs the estimates.

    References:
    
    [1] Yu, Dong, Morten Kolbæk, Zheng-Hua Tan, and Jesper Jensen. 
        "Permutation invariant training of deep models for speaker-independent 
        multi-talker speech separation." In 2017 IEEE International Conference on 
        Acoustics, Speech and Signal Processing (ICASSP), pp. 241-245. IEEE, 2017.
    """
    DEFAULT_KEYS = {'estimates': 'estimates', 'source_magnitudes': 'targets'}

    def __init__(self, loss_function):
        
        super(PermutationInvariantLoss, self).__init__()
        self.loss_function = loss_function
        self.loss_function.reduction = 'none'
        
    def forward(self, estimates, targets):
        num_batch = estimates.shape[0]
        num_sources = estimates.shape[-1]
        estimates = estimates.reshape(num_batch, -1, num_sources)
        targets = targets.reshape(num_batch, -1, num_sources)
        
        losses = []
        for p in permutations(range(num_sources)):
            _targets = targets[..., list(p)]
            loss = self.loss_function(estimates, _targets)
            loss = loss.mean(dim=-1)
            losses.append(loss)
        
        losses = torch.stack(losses, dim=-1)
        losses = torch.min(losses, dim=-1)[0]
        loss = torch.mean(losses)
        return loss


class CombinationInvariantLoss(nn.Module):
    """
    Variant on Permutation Invariant Loss where instead a combination of the
    sources output by the model are used. This way a model can output more 
    sources than there are in the ground truth. A subset of the output sources
    will be compared using Permutation Invariant Loss with the ground truth
    estimates.

    For when you're trying to match the estimates to the sources but you don't 
    know the order in which your model outputs the estimates AND you are 
    outputting more estimates then there are sources.

    """
    DEFAULT_KEYS = {'estimates': 'estimates', 'source_magnitudes': 'targets'}

    def __init__(self, loss_function):
        super(CombinationInvariantLoss, self).__init__()
        self.loss_function = loss_function
        self.loss_function.reduction = 'none'
        
    def forward(self, estimates, targets):
        num_batch = estimates.shape[0]
        num_target_sources = targets.shape[-1]
        num_estimate_sources = estimates.shape[-1]
        
        estimates = estimates.reshape(num_batch, -1, num_estimate_sources)
        targets = targets.reshape(num_batch, -1, num_target_sources)
        
        losses = []
        for c in combinations(range(num_estimate_sources), num_target_sources):
            _estimates = estimates[..., list(c)]
            for p in permutations(range(num_target_sources)):
                _targets = targets[..., list(p)]
                loss = self.loss_function(_estimates, _targets)
                loss = loss.mean(dim=-1)
                losses.append(loss)
        
        losses = torch.stack(losses, dim=-1)
        losses = torch.min(losses, dim=-1)[0]
        loss = torch.mean(losses)
        return loss



import numpy as np

from .. import MaskSeparationBase


class HighLowPassFilter(MaskSeparationBase):
    """
    Implements a super simple separation algorithm that just masks everything below
    the specified hz. It does this by zeroing out the associated FFT bins via a mask to
    produce the "high" source, and the residual is the "low" source.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        high_pass_cutoff_hz (float): Cutoff in Hz. Will be rounded off 
        mask_type (str, optional): Mask type. Defaults to 'binary'.
    """

    def __init__(self, input_audio_signal, high_pass_cutoff_hz, mask_type='binary'):
        super().__init__(input_audio_signal=input_audio_signal, mask_type=mask_type)
        self.high_pass_cutoff_hz = high_pass_cutoff_hz

    def run(self):
        # Compute the spectrogram and find the closest frequency bin to the cutoff freq
        closest_freq_bin = (
            np.abs(self.audio_signal.freq_vector - self.high_pass_cutoff_hz)
        ).argmin()

        # Make masks
        low_pass_mask = self.ones_mask(self.stft.shape)
        low_pass_mask.mask[closest_freq_bin:, ...] = 0
        high_pass_mask = low_pass_mask.invert_mask()
        self.result_masks = [low_pass_mask, high_pass_mask]

        return self.result_masks



import os
import json
import inspect

import torch
from torch import nn
import numpy as np

from . import modules
from ... import __version__


class SeparationModel(nn.Module):
    """
    SeparationModel takes a configuration file or dictionary that describes the model
    structure, which is some combination of MelProjection, Embedding, RecurrentStack,
    ConvolutionalStack, and other modules found in ``nussl.ml.networks.modules``. 

    References:
        Hershey, J. R., Chen, Z., Le Roux, J., & Watanabe, S. (2016, March).
        Deep clustering: Discriminative embeddings for segmentation and separation.
        In Acoustics, Speech and Signal Processing (ICASSP),
        2016 IEEE International Conference on (pp. 31-35). IEEE.

        Luo, Y., Chen, Z., Hershey, J. R., Le Roux, J., & Mesgarani, N. (2017, March).
        Deep clustering and conventional networks for music separation: Stronger together.
        In Acoustics, Speech and Signal Processing (ICASSP),
        2017 IEEE International Conference on (pp. 61-65). IEEE.

    Args:
        config: (str, dict) Either a config dictionary that defines the model and its
          connections, or the path to a json file containing the dictionary. If the
          latter, the path will be loaded and used.

    Attributes:
        config: (dict) The loaded config dictionary passed in upon init.
        connections: (list) A list of strings that define the connections as given
            in `config`.
        output: (list)

    See also:
        ml.register_module to register your custom modules with SeparationModel.

    Examples:
        >>> import nussl
        >>> config = nussl.ml.networks.builders.build_recurrent_dpcl(
        >>>     num_features=512, hidden_size=300, num_layers=3, bidirectional=True,
        >>>     dropout=0.3, embedding_size=20, 
        >>>     embedding_activation=['sigmoid', 'unit_norm'])
        >>>
        >>> model = SeparationModel(config)
    """
    def __init__(self, config, verbose=False):
        super(SeparationModel, self).__init__()
        if type(config) is str:
            if os.path.exists(config):
                with open(config, 'r') as f:
                    config = json.load(f)
            else:
                config = json.loads(config)

        self._validate_config(config)

        module_dict = {}
        self.input = {}
        for module_key in config['modules']:
            module = config['modules'][module_key]

            if 'class' in module:
                if module['class'] in dir(modules): 
                    class_func = getattr(modules, module['class'])
                    module_snapshot = inspect.getsource(class_func)
                else:
                    class_func = getattr(nn, module['class'])
                    module_snapshot = f'pytorch v{torch.__version__} builtin'
                config['modules'][module_key]['module_snapshot'] = module_snapshot

                if 'args' not in module:
                    module['args'] = {}
                module_dict[module_key] = class_func(**module['args'])
            else:
                self.input[module_key] = module_key

        self.layers = nn.ModuleDict(module_dict)
        self.connections = config['connections']
        self.output_keys = config['output']
        self.config = config
        self.verbose = verbose
        self.metadata = {
            'config': config,
            'nussl_version': __version__
        }

    @staticmethod
    def _validate_config(config):
        expected_keys = ['connections', 'modules', 'name', 'output']
        got_keys = sorted(list(config.keys()))

        if got_keys != expected_keys:
            raise ValueError(
                f"Expected keys {expected_keys}, got {got_keys}")

        if not isinstance(config['modules'], dict):
            raise ValueError("config['modules'] must be a dict!")

        if not isinstance(config['connections'], list):
            raise ValueError("config['connections'] must be a list!")

        if not isinstance(config['output'], list):
            raise ValueError("config['output'] must be a list!")

        if not isinstance(config['name'], str):
            raise ValueError("config['name'] must be a string!")

    def forward(self, data):
        """
        Args:
            data: (dict) a dictionary containing the input data for the model. 
            Should match the input_keys in self.input.

        Returns:

        """
        if not all(name in list(data) for name in list(self.input)):
            raise ValueError(
                f'Not all keys present in data! Needs {", ".join(self.input)}')
        output = {}

        for connection in self.connections:
            layer = self.layers[connection[0]]
            input_data = []
            kwargs = {}

            if len(connection) == 2:
                for c in connection[1]:
                    if isinstance(c, dict):
                        for key, val in c.items():
                            if val in output:
                                kwargs[key] = output[val]
                            elif val in data:
                                kwargs[key] = data[val]
                            else:
                                kwargs[key] = val
                    else:
                        input_data.append(output[c] if c in output else data[c])
            _output = layer(*input_data, **kwargs)
            added_keys = []
            if isinstance(_output, dict):
                for k in _output:
                    _key = f'{connection[0]}:{k}'
                    output[_key] = _output[k]
                    added_keys.append(_key)
            elif isinstance(_output, tuple):
                for i, val in enumerate(_output):
                    _key = f'{connection[0]}:{i}'
                    output[_key] = val
                    added_keys.append(_key)
            else:
                _key = connection[0]
                output[_key] = _output
                added_keys.append(_key)
            
            if self.verbose:
                input_shapes = []
                for d in input_data:
                    if torch.is_tensor(d):
                        input_shapes.append(tuple(d.shape))
                input_desc = ", ".join(map(str, input_shapes))
                output_desc = ", ".join(
                    [f"'{k}': {tuple(output[k].shape)}" for k in added_keys])
                stats = {}
                for k in added_keys:
                    stats[k] = {
                        'min': output[k].detach().min().item(),
                        'max': output[k].detach().max().item(),
                        'mean': output[k].detach().mean().item(),
                        'std': output[k].detach().std().item(),
                    }
                
                stats_desc = "\tStatistics:"
                for o in stats:
                    stats_desc += f"\n\t\t{o}"
                    for k in stats[o]: 
                        stats_desc += f"\n\t\t\t{k}: {stats[o][k]:.4f}"

                min_max = ", ".join(
                    [f'{k}: {(output[k].detach().min().item(), output[k].detach().max().item())}' 
                    for k in added_keys]
                )
                print(
                    f"{connection[1]} -> {connection[0]} \n"
                    f"\tTook inputs: {input_desc} \n"
                    f"\tProduced {output_desc} \n"
                    f"{stats_desc}"
                )
                
        return {o: output[o] for o in self.output_keys}

    def save(self, location, metadata=None):
        """
        Saves a SeparationModel into a location into a dictionary with the
        weights and model configuration.
        Args:
            location: (str) Where you want the model saved, as a path.
            metadata: (dict) Additional metadata to save along with the model. By default,
                model config and nussl version is saved as metadata.

        Returns:
            (str): where the model was saved.

        """
        save_dict = {
            'state_dict': self.state_dict(),
            'config': json.dumps(self.config)
        }

        metadata = metadata if metadata else {}
        metadata.update(self.metadata)
        save_dict = {**save_dict, 'metadata': metadata}
        torch.save(save_dict, location)
        return location
    
    def __repr__(self):
        output = super().__repr__()
        num_parameters = 0
        for p in self.parameters():
            if p.requires_grad:
                num_parameters += np.cumprod(p.size())[-1]
        output += '\nNumber of parameters: %d' % num_parameters
        return output
